"""Term-Image's CLI Implementation"""

from __future__ import annotations

import logging as _logging
import os
import sys
import warnings
from contextlib import suppress
from multiprocessing import Event as mp_Event, Queue as mp_Queue, Value
from operator import mul, setitem
from os.path import abspath, basename, exists, isdir, isfile, islink, realpath
from queue import Empty, Queue
from threading import Event, current_thread
from time import sleep
from typing import Any, Callable, Dict, Generator, List, Optional, Tuple, Union
from urllib.parse import urlparse

import PIL
import requests

from . import AutoCellRatio, logging, notify, set_cell_ratio, tui, utils
from .config import config_options, init_config
from .exceptions import StyleError, TermImageError, TermImageWarning, URLNotFoundError
from .exit_codes import FAILURE, INVALID_ARG, NO_VALID_SOURCE, SUCCESS
from .image import BlockImage, ITerm2Image, KittyImage, Size, _best_style
from .image.common import _ALPHA_BG_FORMAT
from .logging import Thread, init_log, log, log_exception
from .logging_multi import Process
from .tui.widgets import Image
from .utils import (
    CSI,
    OS_IS_UNIX,
    clear_queue,
    get_terminal_size,
    set_query_timeout,
    write_tty,
)


def check_dir(
    dir: str, prev_dir: str = "..", *, _links: List[Tuple[str]] = None
) -> Optional[Dict[str, Union[bool, Dict[str, Union[bool, dict]]]]]:
    """Scan *dir* (and sub-directories, if '--recursive' was specified)
    and build the tree of directories [recursively] containing readable images.

    Args:
        - dir: Path to directory to be scanned.
        - prev_dir: Path (absolute or relative to *dir*) to set as working directory
          after scannning *dir* (default:  parent directory of *dir*).
        - _links: Tracks all symlinks from a *source* up **till** a subdirectory.

    Returns:
        - `None` if *dir* contains no readable images [recursively].
        - A dict representing the resulting directory tree whose items are:
          - a "/" key mapped to a ``True``. if *dir* contains image files
          - a directory name mapped to a dict of the same structure, for each non-empty
            sub-directory of *dir*

    NOTE:
        - If '--hidden' was specified, hidden (.[!.]*) images and subdirectories are
          considered.
        - `_depth` should always be initialized, at the module level, before calling
          this function.
    """
    global _depth

    _depth += 1
    try:
        os.chdir(dir)
    except OSError:
        log_exception(
            f"Could not access '{abspath(dir)}{os.sep}'",
            logger,
            direct=True,
        )
        return

    # Some directories can be changed to but cannot be listed
    try:
        entries = os.scandir()
    except OSError:
        log_exception(
            f"Could not get the contents of '{abspath('.')}{os.sep}'",
            logger,
            direct=True,
        )
        os.chdir(prev_dir)
        return

    empty = True
    content = {}
    for entry in entries:
        if interrupted and interrupted.is_set():
            break
        if not SHOW_HIDDEN and entry.name.startswith("."):
            continue
        try:
            is_file = entry.is_file()
            is_dir = entry.is_dir()
        except OSError:
            continue

        if is_file:
            if empty:
                try:
                    PIL.Image.open(entry.name)
                    empty = False
                    if not RECURSIVE:
                        break
                except Exception:
                    pass
        elif RECURSIVE and is_dir:
            if _depth > MAX_DEPTH:
                if not empty:
                    break
                continue

            result = None
            try:
                if entry.is_symlink():
                    path = realpath(entry)

                    # Eliminate cyclic symlinks
                    if os.getcwd().startswith(path) or (
                        _links and any(link[0].startswith(path) for link in _links)
                    ):
                        continue

                    if _source and _free_checkers.value:
                        _dir_queue.put((_source, _links.copy(), abspath(entry), _depth))
                    else:
                        _links.append((abspath(entry), path))
                        del path
                        # Return to the link's parent rather than the linked directory's
                        # parent
                        result = check_dir(entry.name, os.getcwd(), _links=_links)
                        _links.pop()
                else:
                    if _source and _free_checkers.value:
                        _dir_queue.put((_source, _links.copy(), abspath(entry), _depth))
                    else:
                        result = check_dir(entry.name, _links=_links)
            except OSError:
                pass

            if result:
                content[entry.name] = result

    # '/' is an invalid file/directory name on major platforms.
    # On platforms with root directory '/', it can never be the content of a directory.
    if not empty:
        content["/"] = True

    os.chdir(prev_dir)
    _depth -= 1
    return content or None


def check_dirs(
    checker_no: int,
    content_queue: mp_Queue,
    content_updated: mp_Event,
    dir_queue: mp_Queue,
    progress_queue: mp_Queue,
    progress_updated: mp_Event,
    free_checkers: Value,
    globals_: Dict[str, Any],
) -> None:
    """Checks a directory source in a newly **spawned** child process.

    Intended as the *target* of a **spawned** process to parallelize directory checks.
    """
    global _depth, _source

    globals().update(globals_, _free_checkers=free_checkers, _dir_queue=dir_queue)

    NO_CHECK = (None,) * 3
    while True:
        try:
            source, links, subdir, _depth = dir_queue.get_nowait()
        except KeyboardInterrupt:
            progress_queue.put((checker_no, NO_CHECK))
            raise
        except Empty:
            progress_updated.wait()
            progress_queue.put((checker_no, NO_CHECK))
            with free_checkers:
                free_checkers.value += 1
            try:
                source, links, subdir, _depth = dir_queue.get()
            finally:
                with free_checkers:
                    free_checkers.value -= 1

        if not subdir:
            break

        _source = source or subdir
        if not source:
            log(f"Checking {subdir!r}", logger, verbose=True)

        content_path = get_content_path(source, links, subdir)
        if islink(subdir):
            links.append((subdir, realpath(subdir)))
        progress_updated.wait()
        progress_queue.put((checker_no, (source, content_path, _depth)))
        result = None
        try:
            result = check_dir(subdir, _links=links)
        except Exception:
            log_exception(f"Checking {content_path!r} failed", logger, direct=True)
        finally:
            content_updated.wait()
            content_queue.put((source, content_path, result))


def get_content_path(source: str, links: List[Tuple[str]], subdir: str) -> str:
    """Returns the original path from *source* to *subdir*, collapsing all symlinks
    in-between.
    """
    if not (source and links):
        return subdir

    links = iter(links)
    absolute, prev_real = next(links)
    path = source + absolute[len(source) :]
    for absolute, real in links:
        path += absolute[len(prev_real) :]
        prev_real = real
    path += subdir[len(prev_real) :]

    return path


def get_links(source: str, subdir: str) -> List[Tuple[str, str]]:
    """Returns a list of all symlinks (and the directories they point to) between
    *source* and *subdir*.
    """
    if not source:
        return [(subdir, realpath(subdir))] if islink(subdir) else []

    links = [(source, realpath(source))] if islink(source) else []
    # Strips off the basename in case it's a link
    path = os.path.dirname(subdir[len(source) + 1 :])
    if path:
        cwd = os.getcwd()
        os.chdir(source)
        for dir in path.split(os.sep):
            if islink(dir):
                links.append((abspath(dir), realpath(dir)))
            os.chdir(dir)
        os.chdir(cwd)

    return links


def manage_checkers(
    dir_queue: Union[Queue, mp_Queue],
    contents: Dict[str, Union[bool, Dict]],
    images: List[Tuple[str, Generator]],
) -> None:
    """Manages the processing of directory sources in parallel using multiple processes.

    If multiprocessing is not supported on the host platform, the sources are processed
    serially in the current thread of execution, after all file sources have been
    processed.
    """
    global _depth

    def process_result(
        source: str,
        subdir: str,
        result: Union[None, bool, Dict[str, Union[bool, Dict]]],
        n: int = -1,
    ) -> None:
        if n > -1:
            exitcode = -checkers[n].exitcode
            log(
                f"Checker-{n} was terminated "
                + (f"by signal {exitcode} " if exitcode else "")
                + (f"while checking {subdir!r}" if subdir else ""),
                logger,
                _logging.ERROR,
                direct=False,
            )
            if subdir:
                dir_queue.put(
                    (
                        source,
                        get_links(source, subdir),
                        os.path.join(
                            realpath(os.path.dirname(subdir)), basename(subdir)
                        ),
                        result,
                    )
                )
            return

        if result:
            if source not in contents:
                contents[source] = {}
            update_contents(source, contents[source], subdir, result)
        elif not source and subdir not in contents:
            # Marks a potentially empty source
            # If the source is actually empty the dict stays empty
            contents[subdir] = {}

    if logging.MULTI and args.checkers > 1:
        content_queue = mp_Queue()
        content_updated = mp_Event()
        progress_queue = mp_Queue()
        progress_updated = mp_Event()
        free_checkers = Value("i")
        globals_ = {
            name: globals()[name] for name in ("MAX_DEPTH", "RECURSIVE", "SHOW_HIDDEN")
        }

        checkers = [
            Process(
                name=f"Checker-{n}",
                target=check_dirs,
                args=(
                    n,
                    content_queue,
                    content_updated,
                    dir_queue,
                    progress_queue,
                    progress_updated,
                    free_checkers,
                    globals_,
                ),
            )
            for n in range(args.checkers)
        ]

        for checker in checkers:
            checker.start()

        NO_CHECK = (None,) * 3
        try:
            contents[""] = contents
            content_updated.set()
            checks_in_progress = [NO_CHECK] * args.checkers
            progress_updated.set()

            # Wait until at least one checker starts processing a directory
            setitem(checks_in_progress, *progress_queue.get())

            while not (
                interrupted.is_set()  # MainThread has been interrupted
                or not any(checks_in_progress)  # All checkers are dead
                # All checks are done
                or (
                    # No check in progress
                    all(not check or check == NO_CHECK for check in checks_in_progress)
                    # All sources have been passed in
                    and dir_queue.sources_finished
                    # All sources and branched-off subdirectories have been processed
                    and dir_queue.empty()
                    # All progress updates have been processed
                    and progress_queue.empty()
                    # All results have been processed
                    and content_queue.empty()
                )
            ):
                content_updated.clear()
                while not content_queue.empty():
                    process_result(*content_queue.get())
                content_updated.set()

                progress_updated.clear()
                while not progress_queue.empty():
                    setitem(checks_in_progress, *progress_queue.get())
                progress_updated.set()

                for n, checker in enumerate(checkers):
                    if checks_in_progress[n] and not checker.is_alive():
                        # Ensure it's actually the last source processed by the dead
                        # process that's taken into account.
                        progress_updated.clear()
                        while not progress_queue.empty():
                            setitem(checks_in_progress, *progress_queue.get())
                        progress_updated.set()

                        if checks_in_progress[n]:  # Externally terminated
                            process_result(*checks_in_progress[n], n)
                            checks_in_progress[n] = None

                sleep(0.01)  # Allow queue sizes to be updated
        finally:
            if interrupted.is_set():
                clear_queue(dir_queue)
                clear_queue(content_queue)
                clear_queue(progress_queue)
                return

            if not any(checks_in_progress):
                log(
                    "All checkers were terminated, checking directory sources failed!",
                    logger,
                    _logging.ERROR,
                )
                contents.clear()
                return

            for check in checks_in_progress:
                if check:
                    dir_queue.put((None,) * 4)
            for checker in checkers:
                checker.join()
            del contents[""]
            for source, result in tuple(contents.items()):
                if result:
                    images.append((source, ...))
                else:
                    del contents[source]
                    log(f"{source!r} is empty", logger, verbose=True)
    else:
        current_thread.name = "Checker"

        _, links, source, _depth = dir_queue.get()
        while not interrupted.is_set() and source:
            log(f"Checking {source!r}", logger, verbose=True)
            if islink(source):
                links.append((source, realpath(source)))
            result = False
            try:
                result = check_dir(source, os.getcwd(), _links=links)
            except Exception:
                log_exception(f"Checking {source!r} failed", logger, direct=True)
            finally:
                if result:
                    source = abspath(source)
                    contents[source] = result
                    images.append((source, ...))
                elif not interrupted.is_set() and result is None:
                    log(f"{source!r} is empty", logger, verbose=True)
            _, links, source, _depth = dir_queue.get()

        if interrupted.is_set():
            clear_queue(dir_queue)


def update_contents(
    dir: str,
    contents: Dict[str, Union[bool, Dict]],
    subdir: str,
    subcontents: Dict[str, Union[bool, Dict]],
):
    """Updates a directory's content tree with the content tree of a subdirectory."""

    def update_dict(base: dict, update: dict):
        for key in update:
            # "/" can be in *base* if the directory's parent was re-checked
            if key in base and key != "/":
                update_dict(base[key], update[key])
            else:
                base[key] = update[key]

    path = subdir[len(dir) + 1 :].split(os.sep) if dir else [subdir]
    target = path.pop()

    path_iter = iter(path)
    for branch in path_iter:
        try:
            contents = contents[branch]
        except KeyError:
            contents[branch] = {}
            contents = contents[branch]
            break
    for branch in path_iter:
        contents[branch] = {}
        contents = contents[branch]
    if target in contents:
        update_dict(contents[target], subcontents)
    else:
        contents[target] = subcontents


def get_urls(
    url_queue: Queue,
    images: List[Tuple[str, Image]],
    ImageClass: type,
) -> None:
    """Processes URL sources from a/some separate thread(s)"""
    source = url_queue.get()
    while not interrupted.is_set() and source:
        log(f"Getting image from {source!r}", logger, verbose=True)
        try:
            images.append((basename(source), Image(ImageClass.from_url(source))))
        # Also handles `ConnectionTimeout`
        except requests.exceptions.ConnectionError:
            log(f"Unable to get {source!r}", logger, _logging.ERROR)
        except URLNotFoundError as e:
            log(str(e), logger, _logging.ERROR)
        except PIL.UnidentifiedImageError as e:
            log(str(e), logger, _logging.ERROR)
        except Exception:
            log_exception(f"Getting {source!r} failed", logger, direct=True)
        else:
            log(f"Done getting {source!r}", logger, verbose=True)
        source = url_queue.get()

    if interrupted.is_set():
        clear_queue(url_queue)


def open_files(
    file_queue: Queue,
    images: List[Tuple[str, Image]],
    ImageClass: type,
) -> None:
    source = file_queue.get()
    while not interrupted.is_set() and source:
        log(f"Opening {source!r}", logger, verbose=True)
        try:
            images.append((source, Image(ImageClass.from_file(source))))
        except PIL.UnidentifiedImageError as e:
            log(str(e), logger, _logging.ERROR)
        except OSError as e:
            log(f"Could not read {source!r}: {e}", logger, _logging.ERROR)
        except Exception:
            log_exception(f"Opening {source!r} failed", logger, direct=True)
        source = file_queue.get()

    if interrupted.is_set():
        clear_queue(file_queue)


def main() -> None:
    """CLI execution sub-entry-point"""
    from . import config  # Importing a module-level will result in a circular import
    from .parsers import parser, style_parsers

    global args, url_images, MAX_DEPTH, RECURSIVE, SHOW_HIDDEN

    warnings.filterwarnings("error", "", TermImageWarning, "term_image.image.iterm2")

    def check_arg(
        name: str,
        check: Callable[[Any], Any],
        msg: str,
        exceptions: Tuple[Exception] = None,
        *,
        fatal: bool = True,
    ) -> bool:
        """Performs generic argument value checks and outputs the given message if the
        argument value is invalid.

        Returns:
            ``True`` if valid, otherwise ``False``.

        If *exceptions* is :
          - not given or ``None``, the argument is invalid only if ``check(arg)``
            returns a falsy value.
          - given, the argument is invalid if ``check(arg)`` raises one of the given
            exceptions. It's also invalid if it raises any other exception but the
            error message is different.
        """
        value = getattr(args, name)
        if exceptions:
            valid = False
            try:
                check(value)
                valid = True
            except exceptions:
                pass
            except Exception:
                log_exception(
                    "Invalid! See the logs",
                    logger,
                    f"--{name.replace('_', '-')}",
                    direct=True,
                    fatal=fatal,
                )
        else:
            valid = check(value)

        if not valid:
            notify.notify(
                f"{msg} (got: {value!r})",
                notify.CRITICAL if fatal else notify.ERROR,
                f"--{name.replace('_', '-')}",
            )

        return bool(valid)

    args = parser.parse_args()
    MAX_DEPTH = args.max_depth
    RECURSIVE = args.recursive
    SHOW_HIDDEN = args.all

    force_cli_mode = not sys.stdout.isatty() and not args.cli
    if force_cli_mode:
        args.cli = True

    config.user_config_file = args.config
    if args.no_config:
        config.xdg_config_file = None
    init_config()

    # `check_arg()` requires logging.
    init_log(
        (
            args.log_file
            # If the argument is invalid, the error will be emitted later.
            if args.log_file and config_options["log file"].is_valid(args.log_file)
            else config_options.log_file
        ),
        getattr(_logging, args.log_level),
        args.debug,
        config_options.multi if args.multi is None else args.multi,
        args.quiet,
        args.verbose,
        args.verbose_log,
    )

    for details in (
        ("frame_duration", lambda x: x is None or x > 0.0, "must be greater than zero"),
        ("max_depth", lambda x: x > 0, "must be greater than zero"),
        (
            "max_depth",
            lambda x: (
                x + 50 > sys.getrecursionlimit() and sys.setrecursionlimit(x + 50)
            ),
            "too deep",
            (RecursionError, OverflowError),
        ),
        ("repeat", lambda x: x != 0, "must be non-zero"),
        ("alpha", lambda x: 0.0 <= x < 1.0, "out of range"),
        (
            "alpha_bg",
            lambda x: not x or _ALPHA_BG_FORMAT.fullmatch("#" + x),
            "invalid hex color",
        ),
    ):
        if not check_arg(*details):
            return INVALID_ARG

    for name, option in config_options.items():
        var_name = name.replace(" ", "_")
        try:
            arg_value = getattr(args, var_name)
        # Not all config options have corresponding command-line arguments
        except AttributeError:
            continue

        if arg_value is None:
            setattr(args, var_name, option.value)
        elif not option.is_valid(arg_value):
            arg_name = f"--{name.replace(' ', '-')}"
            notify.notify(
                f"{option.error_msg} (got: {arg_value!r})",
                notify.ERROR,
                arg_name,
            )
            option_repr = "null" if option.value is None else repr(option.value)
            notify.notify(
                f"Using config value: {option_repr}",
                context=arg_name,
                verbose=True,
            )
            setattr(args, var_name, option.value)

    set_query_timeout(args.query_timeout)
    utils.SWAP_WIN_SIZE = args.swap_win_size

    if args.auto_cell_ratio:
        args.cell_ratio = None
    try:
        set_cell_ratio(args.cell_ratio or AutoCellRatio.DYNAMIC)
    except TermImageError:
        notify.notify(
            "Auto cell ratio is not supported in the active terminal or on this "
            "platform, using 0.5. It can be set otherwise using `-C | --cell-ratio`.",
            notify.WARNING,
        )
        args.cell_ratio = 0.5

    ImageClass = {
        "auto": None,
        "kitty": KittyImage,
        "iterm2": ITerm2Image,
        "block": BlockImage,
    }[args.style]
    if not ImageClass:
        ImageClass = _best_style()

    if args.force_style or args.style is config_options.style != "auto":
        ImageClass.is_supported()  # Some classes need to set some attributes
        ImageClass._supported = True
    else:
        try:
            ImageClass(None)
        except StyleError:  # Instantiation isn't permitted
            write_tty(f"{CSI}1K\r".encode())  # Erase emitted APCs
            log(
                f"The '{ImageClass}' render style is not supported in the current "
                "terminal! To use it anyways, add '--force-style'.",
                logger,
                _logging.CRITICAL,
            )
            return FAILURE
        except TypeError:  # Instantiation is permitted
            if not ImageClass.is_supported():  # Also sets any required attributes
                write_tty(f"{CSI}1K\r".encode())  # Erase emitted APCs
                log(
                    f"The '{ImageClass}' render style might not be fully supported in "
                    "the current terminal... using it anyways.",
                    logger,
                    _logging.WARNING,
                )

    # Some APCs (e.g kitty's) used for render style support detection get emitted on
    # some non-supporting terminal emulators
    write_tty(f"{CSI}1K\r".encode())  # Erase emitted APCs

    log(f"Using '{ImageClass}' render style", logger, verbose=True)
    style_parser = style_parsers.get(ImageClass.style)
    style_args = vars(style_parser.parse_known_args()[0]) if style_parser else {}

    if ImageClass.style == "iterm2":
        ITerm2Image.JPEG_QUALITY = style_args.pop("jpeg_quality")
        ITerm2Image.NATIVE_ANIM_MAXSIZE = style_args.pop("native_maxsize")
        ITerm2Image.READ_FROM_FILE = style_args.pop("read_from_file")

    try:
        style_args = ImageClass._check_style_args(style_args)
    except ValueError as e:
        notify.notify(str(e), notify.CRITICAL)
        return INVALID_ARG

    if force_cli_mode:
        log(
            "Output is not a terminal, forcing CLI mode!",
            logger,
            _logging.WARNING,
        )

    log("Processing sources...", logger, verbose=True)
    notify.start_loading()

    file_images, url_images, dir_images = [], [], []
    contents = {}
    sources = [
        abspath(source) if exists(source) else source for source in args.sources or "."
    ]
    unique_sources = set()

    url_queue = Queue()
    getters = [
        Thread(
            target=get_urls,
            args=(url_queue, url_images, ImageClass),
            name=f"Getter-{n}",
        )
        for n in range(1, args.getters + 1)
    ]
    getters_started = False

    file_queue = Queue()
    opener = Thread(
        target=open_files,
        args=(file_queue, file_images, ImageClass),
        name="Opener",
    )
    opener_started = False

    if OS_IS_UNIX and not args.cli:
        if args.checkers is None:
            args.checkers = max(
                (
                    len(os.sched_getaffinity(0))
                    if hasattr(os, "sched_getaffinity")
                    else os.cpu_count() or 0
                )
                - 1,
                2,
            )
        dir_queue = mp_Queue() if logging.MULTI and args.checkers > 1 else Queue()
        dir_queue.sources_finished = False
        check_manager = Thread(
            target=manage_checkers,
            args=(dir_queue, contents, dir_images),
            name="CheckManager",
        )
    checkers_started = False

    for source in sources:
        if source in unique_sources:
            log(f"Source repeated: {source!r}", logger, verbose=True)
            continue
        unique_sources.add(source)

        if all(urlparse(source)[:3]):  # Is valid URL
            if not getters_started:
                for getter in getters:
                    getter.start()
                getters_started = True
            url_queue.put(source)
        elif isfile(source):
            if not opener_started:
                opener.start()
                opener_started = True
            file_queue.put(source)
        elif isdir(source):
            if args.cli:
                log(f"Skipping directory {source!r}", logger, verbose=True)
                continue
            if not OS_IS_UNIX:
                dir_images = True
                continue
            if not checkers_started:
                check_manager.start()
                checkers_started = True
            dir_queue.put(("", [], source, 0))
        else:
            log(f"{source!r} is invalid or does not exist", logger, _logging.ERROR)

    # Signal end of sources
    if getters_started:
        for _ in range(args.getters):
            url_queue.put(None)
    if opener_started:
        file_queue.put(None)
    if checkers_started:
        if logging.MULTI and args.checkers > 1:
            dir_queue.sources_finished = True
        else:
            dir_queue.put((None,) * 4)

    interrupt = None
    while True:
        try:
            if getters_started:
                for getter in getters:
                    getter.join()
            if opener_started:
                opener.join()
            if checkers_started:
                check_manager.join()
            break
        except KeyboardInterrupt as e:  # Ensure logs are in correct order
            if not interrupt:  # keep the first
                interrupted.set()
                interrupt = e
    if interrupt:
        raise interrupt from None

    notify.stop_loading()
    while notify.is_loading():
        pass

    if not OS_IS_UNIX and dir_images:
        log(
            "Directory sources skipped, not supported on Windows!",
            logger,
            _logging.ERROR,
        )
        dir_images = []

    log("... Done!", logger, verbose=True)

    images = file_images + url_images + dir_images
    if not images:
        log("No valid source!", logger)
        return NO_VALID_SOURCE

    if args.cli or (
        not args.tui and len(images) == 1 and isinstance(images[0][1], Image)
    ):
        log("Running in CLI mode", logger, direct=False)

        if style_args.get("native") and len(images) > 1:
            style_args["stall_native"] = False

        show_name = len(args.sources) > 1
        for entry in images:
            image = entry[1]._ti_image
            if args.max_pixels_cli and mul(*image._original_size) > args.max_pixels:
                log(
                    f"Has more than the maximum pixel-count, skipping: {entry[0]!r}",
                    logger,
                    _logging.WARNING,
                    verbose=True,
                )
                continue

            if (
                not args.no_anim
                and image._is_animated
                and not style_args.get("native")
                and len(images) > 1
            ):
                log(f"Skipping animated image: {entry[0]!r}", logger, verbose=True)
                continue

            if show_name:
                notify.notify("\n" + basename(entry[0]) + ":")
            try:
                if args.width is None is args.height:
                    args.width = args.auto_size or Size.AUTO
                image.set_size(
                    args.width,
                    args.height,
                    args.h_allow,
                    args.v_allow,
                )
                image.scale = (
                    (args.scale_x, args.scale_y) if args.scale is None else args.scale
                )
                if args.frame_duration:
                    image.frame_duration = args.frame_duration

                if ImageClass.style == "kitty":
                    image.set_render_method(
                        "lines"
                        if (
                            ImageClass._KITTY_VERSION
                            and image._is_animated
                            and not args.no_anim
                        )
                        else "whole"
                    )
                elif ImageClass.style == "iterm2":
                    image.set_render_method(
                        "whole"
                        if (
                            ImageClass._TERM == "konsole"
                            # Always applies to non-native animations also
                            or image.rendered_height <= get_terminal_size()[1]
                        )
                        else "lines"
                    )

                image.draw(
                    *(
                        (None, 1, None, 1)
                        if args.no_align
                        else (
                            args.h_align,
                            args.pad_width,
                            args.v_align,
                            args.pad_height or 1,
                        )
                    ),
                    (
                        None
                        if args.no_alpha
                        else (
                            args.alpha if args.alpha_bg is None else "#" + args.alpha_bg
                        )
                    ),
                    scroll=args.scroll,
                    animate=not args.no_anim,
                    repeat=args.repeat,
                    cached=(
                        not args.cache_no_anim
                        and (args.cache_all_anim or args.anim_cache)
                    ),
                    check_size=not args.oversize,
                    **style_args,
                )

            # Handles `ValueError` and `.exceptions.InvalidSizeError`
            # raised by `BaseImage.set_size()`, scaling value checks
            # or padding width/height checks.
            except (ValueError, StyleError, TermImageWarning) as e:
                notify.notify(str(e), notify.ERROR)
            except BrokenPipeError:
                # Prevent ignored exception message at interpreter shutdown
                with suppress(BrokenPipeError):
                    sys.stdout.close()
                break
    elif OS_IS_UNIX:
        notify.end_loading()
        tui.init(args, style_args, images, contents, ImageClass)
    else:
        log(
            "The TUI is not supported on Windows! Try with `--cli`.",
            logger,
            _logging.CRITICAL,
        )
        return FAILURE

    return SUCCESS


logger = _logging.getLogger(__name__)

# Initially set from within `.__main__.main()`
# Will be updated from `.logging.init_log()` if multiprocessing is enabled
interrupted: Union[None, Event, mp_Event] = None

# The annotations below are put in comments for compatibility with Python 3.7
# as it doesn't allow names declared as `global` within functions to be annotated.

# Used by `check_dir()`
_depth = None  #: int

# Set from within `check_dirs()`; Hence, only set in "Checker-?" processes
_dir_queue = None  #: Union[None, Queue, mp_Queue]
_free_checkers = None  #: Optional[Value]
_source = None  #: Optional[str]

# Set from within `main()`
MAX_DEPTH = None  #: Optional[int]
RECURSIVE = None  #: Optional[bool]
SHOW_HIDDEN = None  #: Optional[bool]
# # Used in other modules
args = None  #: Optional[argparse.Namespace]
url_images = None  #: Optional[list]



"""Extension of `.logging` for multiprocessing support"""

from __future__ import annotations

import logging as _logging
import os
from multiprocessing import JoinableQueue, Process
from traceback import format_exception

from . import AutoCellRatio, cli, logging, notify, set_cell_ratio, tui, utils


def process_multi_logs() -> None:
    """Emits logs and notifications redirected from subprocesses.

    Intended to be executed in a separate thread of the main process.
    """
    global log_queue

    PID = os.getpid()
    log_queue = JoinableQueue()
    process_multi_logs.started.set()

    log_type, data = log_queue.get()
    while data:
        if log_type == LOG:
            data["process"] = PID
            _logger.handle(_logging.makeLogRecord(data))
        else:
            notify.notify(*data[0], **data[1])
        log_queue.task_done()
        log_type, data = log_queue.get()
    log_queue.task_done()


class Process(Process):
    """A process with integration into the logging system

    Sets up the logging system to redirect all logs (and optionally notifcations)
    in the subprocess, to the main process to be emitted.

    NOTE:
        - Only TUI notifications need to be redirected.
        - The redirected logs and notifcations are automatically handled by
          `process_multi_logs()`, running in the MultiLogger thread of the main process.
    """

    def __init__(self, *args, redirect_notifs: bool = False, **kwargs):
        super().__init__(*args, **kwargs)
        self._log_queue = log_queue
        self._logging_details = {
            "constants": {
                name: value for name, value in vars(logging).items() if name.isupper()
            },
            "logging_level": _logging.getLogger().getEffectiveLevel(),
            "redirect_notifs": redirect_notifs,
        }
        self._main_process_interrupted = cli.interrupted
        self._ImageClass = tui.main.ImageClass
        if self._ImageClass:  # if the TUI is initialized
            self._cell_ratio = cli.args.cell_ratio
            self._query_timeout = utils.QUERY_TIMEOUT
            self._swap_win_size = utils.SWAP_WIN_SIZE
            self._style_attrs = [
                (attr, getattr(self._ImageClass, attr))
                for attr in exported_style_attrs.get(self._ImageClass.style, ())
            ]
        child_processes.append(self)

    def run(self):
        self._redirect_logs()
        _logger.debug("Starting")

        try:
            if self._ImageClass:  # if the TUI is initialized
                # The unpickled class object is in the originally defined state
                # Eliminates queries for style support checks
                self._ImageClass._supported = True
                for item in self._style_attrs:
                    setattr(self._ImageClass, *item)

                utils.QUERY_TIMEOUT = self._query_timeout
                utils.SWAP_WIN_SIZE = self._swap_win_size

                if not self._cell_ratio:
                    # Avoid an error in case the terminal wouldn`t respond on time
                    AutoCellRatio.is_supported = True
                set_cell_ratio(self._cell_ratio or AutoCellRatio.DYNAMIC)

            super().run(set_tty_lock=bool(self._ImageClass))
        except KeyboardInterrupt:
            # Log only if the main process was not interrupted
            if not self._main_process_interrupted.wait(0.1):
                logging.log(
                    "Interrupted" if logging.DEBUG else f"{self.name} was interrupted",
                    _logger,
                    _logging.ERROR,
                    direct=False,
                )
        except Exception:
            logging.log_exception(
                "Aborted" if logging.DEBUG else f"{self.name} was aborted", _logger
            )
        else:
            _logger.debug("Exiting")

    def _notif_redirector(self, *args, loading: bool = False, **kwargs):
        self._log_queue.put((NOTIF, (args, kwargs)))

    def _redirect_logs(self) -> None:
        # Logs
        vars(logging).update(self._logging_details["constants"])
        logger = _logging.getLogger()
        logger.setLevel(self._logging_details["logging_level"])
        logger.addHandler(RedirectHandler(self._log_queue))
        logger.handlers[0].addFilter(logging.filter_)

        # # Warnings and session-level logs
        _logger.setLevel(min(self._logging_details["logging_level"], _logging.INFO))

        # Notifications
        if self._logging_details["redirect_notifs"] and not logging.QUIET:
            notify.notify = self._notif_redirector


class RedirectHandler(_logging.Handler):
    """Puts the attribute dict of log records into *log_queue*.

    The records can be recreated with `logging.makeLogRecord()` and emitted with the
    `handle()` method of a logger with a different handler.
    """

    def __init__(self, log_queue: JoinableQueue, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._log_queue = log_queue

    def handle(self, record: _logging.LogRecord):
        attrdict = vars(record)
        exc_info = attrdict["exc_info"]
        if exc_info:
            # traceback objects cannot be pickled
            attrdict["msg"] = "\n".join(
                (attrdict["msg"], "".join(format_exception(*exc_info)))
            ).rstrip()
            attrdict["exc_info"] = None
        self._log_queue.put((LOG, attrdict))


_logger = _logging.getLogger("term-image")

LOG = 0
NOTIF = 1
child_processes = []
exported_style_attrs = {
    "iterm2": ("_TERM", "_TERM_VERSION"),
    "kitty": ("_TERM", "_TERM_VERSION", "_KITTY_VERSION"),
}

# The annotations below are put in comments for compatibility with Python 3.7
# as it doesn't allow names declared as `global` within functions to be annotated.

# Set from `process_multi_logs()` in the MultiLogger thread, only in the main process
log_queue = None  #: Optional[JoinableQueue]



"""
.. Common Interfaces For Various Image Classes
"""

from __future__ import annotations

__all__ = (
    "ImageSource",
    "Size",
    "ImageMeta",
    "BaseImage",
    "GraphicsImage",
    "TextImage",
    "ImageIterator",
)

import io
import os
import re
import sys
import time
from abc import ABCMeta, abstractmethod
from enum import Enum
from functools import wraps
from math import ceil
from operator import gt, mul, sub
from random import randint
from types import FunctionType, TracebackType
from typing import Any, Dict, List, Optional, Set, Tuple, Union
from urllib.parse import urlparse

import PIL
import requests
from PIL import Image, UnidentifiedImageError

from .. import get_cell_ratio
from ..exceptions import (
    InvalidSizeError,
    TermImageError,
    URLNotFoundError,
    _style_error,
)
from ..utils import (
    COLOR_RESET,
    CSI,
    ClassInstanceMethod,
    cached,
    get_cell_size,
    get_fg_bg_colors,
    get_terminal_name_version,
    get_terminal_size,
    no_redecorate,
)

_ALPHA_THRESHOLD = 40 / 255  # Default alpha threshold
_FORMAT_SPEC = re.compile(
    r"(([<|>])?(\d+)?)?(\.([-^_])?(\d+)?)?(#(\.\d+|[0-9a-f]{6}|#)?)?(\+(.+))?",
    re.ASCII,
)
_NO_VERTICAL_SPEC = re.compile(r"(([<|>])?(\d+)?)?\.(#(\.\d+|[0-9a-f]{6})?)?", re.ASCII)
_ALPHA_BG_FORMAT = re.compile("#([0-9a-f]{6})?", re.ASCII)


@no_redecorate
def _close_validated(func: FunctionType) -> FunctionType:
    """Enables finalization status validation before performing an operation with a
    `BaseImage` instance.
    """

    @wraps(func)
    def close_validated_wrapper(self, *args, **kwargs):
        if self._closed:
            raise TermImageError("This image has been finalized")
        return func(self, *args, **kwargs)

    return close_validated_wrapper


class Hidden:
    """An object that hides it's original value representation."""

    def __repr__(_):
        return "<hidden>"

    __ascii__ = __str__ = __repr__


class ImageSource(Enum):
    """Image source type.

    NOTE:
        The values of the enumeration members are implementation details and might
        change at anytime.
        Any comparison should be by identity of the members themselves.
    """

    _ignore_ = ["_SourceAttr"]

    class _SourceAttr(Hidden, str):
        """A string that only compares equal to itself but returns the original hash of
        the string.

        Used to store the attribute that holds the value for ``image.source`` as the
        value of enum members, because some would normally compare equal.
        """

        def __init__(self, *_):
            self._str = super(Hidden).__str__()

        def __eq__(self, _):
            return NotImplemented

        __ne__ = __eq__
        __hash__ = str.__hash__

    #: The instance was derived from a path to a local image file.
    FILE_PATH = _SourceAttr("_source")

    #: The instance was derived from a PIL image instance.
    PIL_IMAGE = _SourceAttr("_source")

    #: The instance was derived from an image URL.
    URL = _SourceAttr("_url")


class Size(Enum):
    """Enumeration for :term:`automatic sizing`"""

    #: Equivalent to :py:attr:`ORIGINAL` if it will fit into the
    #: :term:`available size`, else :py:attr:`FIT`.
    AUTO = Hidden()

    #: The image size is set to fit optimally **within** the :term:`available size`.
    FIT = Hidden()

    #: The size is set such that the width is exactly the :term:`available width`,
    #: regardless of the :term:`cell ratio`.
    FIT_TO_WIDTH = Hidden()

    #: The image size is set such that the image is rendered with as many pixels as the
    #: the original image consists of.
    ORIGINAL = Hidden()


class ImageMeta(ABCMeta):
    """Type of all render style classes.

    NOTE:
        | For all render style classes (instances of this class) defined **within** this
          package, ``str(cls)`` yeilds the same value as :py:attr:`cls.style <style>`.
        | For render style classes defined **outside** this package (subclasses of those
          defined within this package), ``str(cls)`` is equivalent to ``repr(cls)``.
    """

    def __str__(self):
        return self.style or super().__str__()

    style = property(
        cached(
            lambda self: (
                self.__name__[:-5].lower()
                if self.__module__.startswith("term_image.image")
                else None
            )
        ),
        doc="""Name of the render style [category].

        Returns:
            * The name of the render style [category] implemented by the invoking
              class, if defined **within** this package (``term_image``)
            * ``None``, if the invoking class is defined **outside** this package
              (``term_image``)

        :rtype: Optional[str]

        **Examples**

        For a class defined within this package:

        >>> from term_image.image import KittyImage
        >>> KittyImage.style
        'kitty'

        For a class defined outside this package:

        >>> from term_image.image import KittyImage
        >>> class MyImage(KittyImage): pass
        >>> MyImage.style is None
        True

        HINT:
            Equivalent to ``str(cls)`` for all render style classes (instances of
            :py:class:`ImageMeta`) defined **within** this package.
        """,
    )


class BaseImage(metaclass=ImageMeta):
    """Base of all render styles.

    Args:
        image: Source image.
        width: Can be

          * an ``int``; horizontal dimension of the image, in columns.
          * a :py:class:`~term_image.image.Size` enum member.

        height: Can be

          * an ``int``; vertical dimension of the image, in lines.
          * a :py:class:`~term_image.image.Size` enum member.

        scale: The fraction of the size (on respective axes) to render the image with.

    Raises:
        TypeError: An argument is of an inappropriate type.
        ValueError: An argument is of an appropriate type but has an
          unexpected/invalid value.

    Propagates exceptions raised by :py:meth:`set_size`, if *width* or *height* is
    given.

    NOTE:
        * If neither *width* nor *height* is given (or both are ``None``),
          :py:attr:`~term_image.image.Size.FIT` applies.
        * The image size is multiplied by the :term:`scale` on respective axes when
          the image is :term:`rendered`.
        * For animated images, the seek position is initialized to the current seek
          position of the given image.

    ATTENTION:
        This class cannot be directly instantiated. Image instances should be created
        from its subclasses.
    """

    # Data Attributes

    _supported: Optional[bool] = None
    _render_method: Optional[str] = None
    _render_methods: Set[str] = set()
    _style_args: Dict[
        str, Tuple[Tuple[FunctionType, str], Tuple[FunctionType, str]]
    ] = {}

    # Special Methods

    def __init__(
        self,
        image: PIL.Image.Image,
        *,
        width: Union[int, Size, None] = None,
        height: Union[int, Size, None] = None,
        scale: Tuple[float, float] = (1.0, 1.0),
    ) -> None:
        """See the class description"""
        if not isinstance(image, Image.Image):
            raise TypeError(
                "Expected a 'PIL.Image.Image' instance for 'image' "
                f"(got: {type(image).__name__!r})."
            )

        self._closed = False
        self._source = image
        self._source_type = ImageSource.PIL_IMAGE
        self._original_size = image.size
        if width is None is height:
            self.size = Size.FIT
        else:
            self.set_size(width, height)
        self._scale = []
        self._scale[:] = self._check_scale(scale)

        self._is_animated = hasattr(image, "is_animated") and image.is_animated
        if self._is_animated:
            self._frame_duration = (image.info.get("duration") or 100) / 1000
            self._seek_position = image.tell()
            self._n_frames = None

    def __del__(self) -> None:
        self.close()

    def __enter__(self) -> BaseImage:
        return self

    def __exit__(self, typ: type, val: Exception, tb: TracebackType) -> bool:
        self.close()
        return False  # Currently, no particular exception is suppressed

    def __format__(self, spec: str) -> str:
        """Renders the image with alignment, padding and transparency control"""
        # Only the currently set frame is rendered for animated images
        h_align, width, v_align, height, alpha, style_args = self._check_format_spec(
            spec
        )

        return self._format_render(
            self._renderer(self._render_image, alpha, **style_args),
            h_align,
            width,
            v_align,
            height,
        )

    def __iter__(self) -> ImageIterator:
        return ImageIterator(self, 1, "1.1", False)

    def __repr__(self) -> str:
        return "<{}: source_type={} size={} scale={} is_animated={}>".format(
            type(self).__name__,
            self._source_type.name,
            (
                self._size.name
                if isinstance(self._size, Size)
                else "x".join(map(str, self._size))
            ),
            "x".join(format(x, ".2") for x in self._scale),
            self._is_animated,
        )

    def __str__(self) -> str:
        """Renders the image with transparency enabled and without alignment"""
        # Only the currently set frame is rendered for animated images
        return self._renderer(self._render_image, _ALPHA_THRESHOLD)

    # Properties

    closed = property(
        lambda self: self._closed,
        doc="""Instance finalization status

        :rtype: bool
        """,
    )

    frame_duration = property(
        lambda self: self._frame_duration if self._is_animated else None,
        doc="""Duration (in seconds) of a single frame for :term:`animated` images

        Setting this on non-animated images is simply ignored, no exception is raised.

        :rtype: float
        """,
    )

    @frame_duration.setter
    def frame_duration(self, value: float) -> None:
        if not isinstance(value, float):
            raise TypeError(f"Invalid duration type (got: {type(value).__name__})")
        if value <= 0.0:
            raise ValueError(f"Invalid frame duration (got: {value})")
        if self._is_animated:
            self._frame_duration = value

    height = property(
        lambda self: self._size if isinstance(self._size, Size) else self._size[1],
        lambda self, height: self.set_size(height=height),
        doc="""
        The **unscaled** height of the image.

        Returns:

          * The image height (in lines), if the image size is
            :term:`fixed <fixed size>`.
          * A :py:class:`~term_image.image.Size` enum member; if the image size
            is :term:`dynamic <dynamic size>`.

        :rtype: Union[Size, int]

        SETTABLE VALUES:

        * a positive ``int``; the image height is set to the given value and the
          width is set proportionally.
        * a :py:class:`~term_image.image.Size` enum member; the image size
          is set as prescibed by the enum member.
        * ``None``; equivalent to :py:attr:`~term_image.image.Size.FIT`.

        Setting this

        * results in a :term:`fixed size`.
        * resets the recognized advanced sizing options to their defaults.
        """,
    )

    is_animated = property(
        lambda self: self._is_animated,
        doc="``True`` if the image is :term:`animated`. Otherwise, ``False``.",
    )

    original_size = property(
        lambda self: self._original_size, doc="Size of the source image (in pixels)"
    )

    @property
    def n_frames(self) -> int:
        """The number of frames in the image

        :rtype: int
        """
        if not self._is_animated:
            return 1

        if not self._n_frames:
            self._n_frames = self._get_image().n_frames

        return self._n_frames

    rendered_height = property(
        lambda self: round(
            (
                self._valid_size(None, self._size)
                if isinstance(self._size, Size)
                else self._size
            )[1]
            * self._scale[1]
        ),
        doc="""
        The **scaled** height of the image.

        Also the exact number of lines that the drawn image will occupy in a terminal.

        :rtype: int
        """,
    )

    rendered_size = property(
        lambda self: tuple(
            map(
                round,
                map(
                    mul,
                    (
                        self._valid_size(self._size, None)
                        if isinstance(self._size, Size)
                        else self._size
                    ),
                    self._scale,
                ),
            )
        ),
        doc="""
        The **scaled** size of the image.

        Also the exact number of columns and lines (respectively) that the drawn image
        will occupy in a terminal.

        :rtype: Tuple[int, int]
        """,
    )

    rendered_width = property(
        lambda self: round(
            (
                self._valid_size(self._size, None)
                if isinstance(self._size, Size)
                else self._size
            )[0]
            * self._scale[0]
        ),
        doc="""
        The **scaled** width of the image.

        Also the exact number of columns that the drawn image will occupy in a terminal.

        :rtype: int
        """,
    )

    scale = property(
        lambda self: tuple(self._scale),
        doc="""
        Image :term:`scale`

        SETTABLE VALUES:

          * A *scale value*; sets both axes.
          * A ``tuple`` of two *scale values*; sets ``(x, y)`` respectively.

        A scale value is a ``float`` in the range **0.0 < value <= 1.0**.

        :rtype: Tuple[float, float]
        """,
    )

    @scale.setter
    def scale(self, scale: Union[float, Tuple[float, float]]) -> None:
        if isinstance(scale, float):
            if not 0.0 < scale <= 1.0:
                raise ValueError(f"Scale value out of range (got: {scale})")
            self._scale[:] = (scale,) * 2
        elif isinstance(scale, tuple):
            self._scale[:] = self._check_scale(scale)
        else:
            raise TypeError("'scale' must be a float or a tuple of floats")

    scale_x = property(
        lambda self: self._scale[0],
        doc="""
        Horizontal :term:`scale`

        A scale value is a ``float`` in the range **0.0 < x <= 1.0**.

        :rtype: float
        """,
    )

    @scale_x.setter
    def scale_x(self, x: float) -> None:
        self._scale[0] = self._check_scale_2(x)

    scale_y = property(
        lambda self: self._scale[1],
        doc="""
        Vertical :term:`scale`

        A scale value is a ``float`` in the range **0.0 < y <= 1.0**.

        :rtype: float
        """,
    )

    @scale_y.setter
    def scale_y(self, y: float) -> None:
        self._scale[1] = self._check_scale_2(y)

    size = property(
        lambda self: self._size,
        doc="""
        The **unscaled** size of the image.

        Returns:

          * The image size, ``(columns, lines)``, if the image size is
            :term:`fixed <fixed size>`.
          * A :py:class:`~term_image.image.Size` enum member, if the image size
            is :term:`dynamic <dynamic size>`.

        :rtype: Union[Size, Tuple[int, int]]

        SETTABLE VALUES:

        * A :py:class:`~term_image.image.Size` enum member; the image size
          is set as prescibed by the enum member.

        Setting this

        * implies :term:`dynamic sizing` i.e the size is computed whenever the image is
          :term:`rendered`.
        * resets the recognized advanced sizing options to their defaults.

        This is multiplied by the :term:`scale` on respective axes when the image
        is :term:`rendered`.
        """,
    )

    @size.setter
    def size(self, value: Size) -> None:
        if not isinstance(value, Size):
            raise TypeError("'size' must be a `Size` enum member")
        self._size = value
        self._fit_to_width = value is Size.FIT_TO_WIDTH
        self._h_allow = 0
        self._v_allow = 2  # A 2-line allowance for the shell prompt, etc

    source = property(
        _close_validated(lambda self: getattr(self, self._source_type.value)),
        doc="""
        The :term:`source` from which the instance was initialized.

        :rtype: Union[PIL.Image.Image, str]
        """,
    )

    source_type = property(
        lambda self: self._source_type,
        doc="""
        The kind of :term:`source` from which the instance was initialized.

        :rtype: ImageSource
        """,
    )

    width = property(
        lambda self: self._size if isinstance(self._size, Size) else self._size[0],
        lambda self, width: self.set_size(width),
        doc="""
        The **unscaled** width of the image.

        Returns:

          * The image width (in columns), if the image size is
            :term:`fixed <fixed size>`.
          * A :py:class:`~term_image.image.Size` enum member; if the image size
            is :term:`dynamic <dynamic size>`.

        :rtype: Union[Size, int]

        SETTABLE VALUES:

        * a positive ``int``; the image width is set to the given value and the
          height is set proportionally.
        * a :py:class:`~term_image.image.Size` enum member; the image size
          is set as prescibed by the enum member.
        * ``None``; equivalent to :py:attr:`~term_image.image.Size.FIT`.

        Setting this

        * results in a :term:`fixed size`.
        * resets the recognized advanced sizing options to their defaults.
        """,
    )

    # # Private

    @property
    @abstractmethod
    def _pixel_ratio(self):
        """The width-to-height ratio of a pixel drawn in the terminal"""
        raise NotImplementedError

    # Public Methods

    def close(self) -> None:
        """Finalizes the instance and releases external resources.

        * In most cases, it's not neccesary to explicity call this method, as it's
          automatically called when the instance is garbage-collected.
        * This method can be safely called mutiple times.
        * If the instance was initialized with a PIL image, the PIL image is never
          finalized.
        """
        try:
            if not self._closed:
                if self._source_type is ImageSource.URL:
                    try:
                        os.remove(self._source)
                    except FileNotFoundError:
                        pass
                    del self._url
                del self._source
        except AttributeError:
            pass  # Instance creation or initialization was unsuccessful
        finally:
            self._closed = True

    def draw(
        self,
        h_align: Optional[str] = None,
        pad_width: Optional[int] = None,
        v_align: Optional[str] = None,
        pad_height: Optional[int] = None,
        alpha: Optional[float, str] = _ALPHA_THRESHOLD,
        *,
        scroll: bool = False,
        animate: bool = True,
        repeat: int = -1,
        cached: Union[bool, int] = 100,
        check_size: bool = True,
        **style: Any,
    ) -> None:
        """Draws an image to standard output.

        Args:
            h_align: Horizontal alignment ("left" / "<", "center" / "|" or
              "right" / ">"). Default: center.
            pad_width: Number of columns within which to align the image.

              * Excess columns are filled with spaces.
              * Must not be greater than the
                :term:`available terminal width <available width>`.
              * Default: terminal width, minus horizontal allowance.

            v_align: Vertical alignment ("top"/"^", "middle"/"-" or "bottom"/"_").
              Default: middle.
            pad_height: Number of lines within which to align the image.

              * Excess lines are filled with spaces.
              * Must not be greater than the :term:`available terminal height
                <available height>`, **for animations**.
              * Default: terminal height, minus vertical allowance.

            alpha: Transparency setting.

              * If ``None``, transparency is disabled (alpha channel is removed).
              * If a ``float`` (**0.0 <= x < 1.0**), specifies the alpha ratio
                **above** which pixels are taken as **opaque**. **(Applies to only
                text-based render styles)**.
              * If a string, specifies a color to replace transparent background with.
                Can be:

                * **"#"** -> The terminal's default background color (or black, if
                  undetermined) is used.
                * A hex color e.g ``ffffff``, ``7faa52``.

            scroll: Only applies to non-animations. If ``True``, allows the image's
                :term:`rendered height` to be greater than the
                :term:`available terminal height <available height>`.

            animate: If ``False``, disable animation i.e draw only the current frame of
              an animated image.
            repeat: The number of times to go over all frames of an animated image.
              A negative value implies infinite repetition.
            cached: Determines if :term:`rendered` frames of an animated image will be
              cached (for speed up of subsequent renders of the same frame) or not.

              * If ``bool``, it directly sets if the frames will be cached or not.
              * If ``int``, caching is enabled only if the framecount of the image
                is less than or equal to the given number.

            check_size: If ``False``, does not perform size validation for
              non-animations.
            style: Style-specific parameters. See each subclass for it's own usage.

        Raises:
            TypeError: An argument is of an inappropriate type.
            ValueError: An argument is of an appropriate type but has an
              unexpected/invalid value.
            ValueError: Unable to convert image.
            ValueError: Image size or :term:`scale` too small.
            term_image.exceptions.InvalidSizeError: The image's :term:`rendered size`
              can not fit into the :term:`available terminal size <available size>`.
            term_image.exceptions.StyleError: Unrecognized style-specific parameter(s).

        * If :py:meth:`set_size` was used to set the image size, the horizontal and
          vertical allowances (set when :py:meth:`set_size` was called) are taken into
          consideration during size validation. If the size was set via another means or
          the size is :term:`dynamic <dynamic size>`, the default allowances apply.
        * For **non-animations**, if the image size was set with
          :py:attr:term_image.image.Size.FIT_TO_WIDTH`, the image **height** is not
          validated and setting *scroll* is unnecessary.
        * *animate*, *repeat* and *cached* apply to :term:`animated` images only.
          They are simply ignored for non-animated images.
        * For animations (i.e animated images with *animate* set to ``True``):

          * *scroll* is ignored.
          * Image size and :term:`padding height` are always validated, if set or given.
          * **with the exception of native animations provided by some render styles**.

        * Animations, **by default**, are infinitely looped and can be terminated
          with **Ctrl+C** (``SIGINT``), raising ``KeyboardInterrupt``.
        """
        fmt = self._check_formatting(h_align, pad_width, v_align, pad_height)

        if alpha is not None:
            if isinstance(alpha, float):
                if not 0.0 <= alpha < 1.0:
                    raise ValueError(f"Alpha threshold out of range (got: {alpha})")
            elif isinstance(alpha, str):
                if not _ALPHA_BG_FORMAT.fullmatch(alpha):
                    raise ValueError(f"Invalid hex color string (got: {alpha})")
            else:
                raise TypeError(
                    "'alpha' must be `None` or of type `float` or `str` "
                    f"(got: {type(alpha).__name__})"
                )

        if self._is_animated and not isinstance(animate, bool):
            raise TypeError("'animate' must be a boolean")

        if None is not pad_width > get_terminal_size()[0] - self._h_allow:
            raise ValueError(
                "Padding width is greater than the available terminal width"
            )

        if (
            not style.get("native")
            and self._is_animated
            and animate
            and None is not pad_height > get_terminal_size()[1]
        ):
            raise ValueError(
                "Padding height can not be greater than the terminal height for "
                "animations"
            )

        for arg in ("scroll", "check_size"):
            if not isinstance(locals()[arg], bool):
                raise TypeError(f"{arg!r} must be a boolean")

        # Checks for *repeat* and *cached* are delegated to `ImageIterator`.

        def render(image: PIL.Image.Image) -> None:
            # Hide the cursor immediately if the output is a terminal device
            sys.stdout.isatty() and print(f"{CSI}?25l", end="", flush=True)
            try:
                style_args = self._check_style_args(style)
                if self._is_animated and animate:
                    self._display_animated(
                        image, alpha, fmt, repeat, cached, **style_args
                    )
                else:
                    try:
                        print(
                            self._format_render(
                                self._render_image(image, alpha, **style_args),
                                *fmt,
                            ),
                            end="",
                            flush=True,
                        )
                    except (KeyboardInterrupt, Exception):
                        self._handle_interrupted_draw()
                        raise
            finally:
                # Reset color and show the cursor
                print(COLOR_RESET, f"{CSI}?25h" * sys.stdout.isatty(), sep="")

        self._renderer(
            render,
            scroll=scroll,
            check_size=check_size,
            animated=not style.get("native") and self._is_animated and animate,
        )

    @classmethod
    def from_file(
        cls,
        filepath: str,
        **kwargs: Union[None, int, Tuple[float, float]],
    ) -> BaseImage:
        """Creates an instance from an image file.

        Args:
            filepath: Relative/Absolute path to an image file.
            kwargs: Same keyword arguments as the class constructor.

        Returns:
            A new instance.

        Raises:
            TypeError: *filepath* is not a string.
            FileNotFoundError: The given path does not exist.
            IsADirectoryError: Propagated from from ``PIL.Image.open()``.
            PIL.UnidentifiedImageError: Propagated from from ``PIL.Image.open()``.

        Also Propagates exceptions raised or propagated by the class constructor.
        """
        if not isinstance(filepath, str):
            raise TypeError(
                f"File path must be a string (got: {type(filepath).__name__!r})."
            )

        # Intentionally propagates `IsADirectoryError` since the message is OK
        try:
            new = cls(Image.open(filepath), **kwargs)
        except FileNotFoundError:
            raise FileNotFoundError(f"No such file: {filepath!r}") from None
        except UnidentifiedImageError as e:
            e.args = (f"Could not identify {filepath!r} as an image",)
            raise

        # Absolute paths work better with symlinks, as opposed to real paths:
        # less confusing, Filename is as expected, helps in path comparisons
        new._source = os.path.abspath(filepath)
        new._source_type = ImageSource.FILE_PATH
        return new

    @classmethod
    def from_url(
        cls,
        url: str,
        **kwargs: Union[None, int, Tuple[float, float]],
    ) -> BaseImage:
        """Creates an instance from an image URL.

        Args:
            url: URL of an image file.
            kwargs: Same keyword arguments as the class constructor.

        Returns:
            A new instance.

        Raises:
            TypeError: *url* is not a string.
            ValueError: The URL is invalid.
            term_image.exceptions.URLNotFoundError: The URL does not exist.
            PIL.UnidentifiedImageError: Propagated from ``PIL.Image.open()``.

        Also propagates connection-related exceptions from ``requests.get()``
        and exceptions raised or propagated by the class constructor.

        NOTE:
            This method creates a temporary image file, but only after a successful
            initialization.

            Proper clean-up is guaranteed except maybe in very rare cases.

            To ensure 100% guarantee of clean-up, use the object as a
            :ref:`context manager <context-manager>`.
        """
        if not isinstance(url, str):
            raise TypeError(f"URL must be a string (got: {type(url).__name__!r}).")
        if not all(urlparse(url)[:3]):
            raise ValueError(f"Invalid URL: {url!r}")

        # Propagates connection-related errors.
        response = requests.get(url, stream=True)
        if response.status_code == 404:
            raise URLNotFoundError(f"URL {url!r} does not exist.")

        try:
            new = cls(Image.open(io.BytesIO(response.content)), **kwargs)
        except UnidentifiedImageError as e:
            e.args = (f"The URL {url!r} doesn't link to an identifiable image",)
            raise

        # Ensure initialization is successful before writing to file

        basedir = os.path.join(os.path.expanduser("~"), ".term_image", "temp")
        if not os.path.isdir(basedir):
            os.makedirs(basedir)

        filepath = os.path.join(basedir, os.path.basename(url))
        while os.path.exists(filepath):
            filepath += str(randint(0, 9))
        with open(filepath, "wb") as image_writer:
            image_writer.write(response.content)

        new._source = filepath
        new._source_type = ImageSource.URL
        new._url = url
        return new

    @classmethod
    @abstractmethod
    def is_supported(cls) -> bool:
        """Returns ``True`` if the render style or graphics protocol implemented by
        the invoking class is supported by the :term:`active terminal`.
        Otherwise, ``False``.

        ATTENTION:
            Support checks for most (if not all) render styles require :ref:`querying
            <terminal-queries>` the :term:`active terminal`, though **only the first
            time** they're executed.

            Hence, it's advisable to perform all neccesary support checks (call
            ``is_supported()`` on required subclasses) at an early stage of a program,
            before user input is required.
        """
        raise NotImplementedError

    def seek(self, pos: int) -> None:
        """Changes current image frame.

        Args:
            pos: New frame number.

        Raises:
            TypeError: An argument is of an inappropriate type.
            ValueError: An argument is of an appropriate type but has an
              unexpected/invalid value.

        Frame numbers start from 0 (zero).
        """
        if not isinstance(pos, int):
            raise TypeError(f"Invalid frame number type (got: {type(pos).__name__})")
        if not 0 <= pos < self.n_frames:
            raise ValueError(
                f"Frame number out of range (got: {pos}, n_frames={self.n_frames})"
            )
        if self._is_animated:
            self._seek_position = pos

    @ClassInstanceMethod
    def set_render_method(self_or_cls, method: Optional[str] = None) -> None:
        """Sets the render method used by the instances of subclasses providing
        multiple render methods.

        Args:
            method: The render method to be set or ``None`` for a reset
              (case-insensitive).

        Raises:
            TypeError: *method* is not a string or ``None``.
            ValueError: the given method is not implmented by the invoking class
              (or class of the invoking instance).

        See the **Render Methods** section in the description of the subclasses that
        implement such for their specific usage.

        If called via:

           - a class, sets the class-wide render method.
           - an instance, sets the instance-specific render method.

        If *method* is ``None`` and this method is called via:

           - a class, the class-wide render method is reset to the default.
           - an instance, the instance-specific render method is removed, so that it
             uses the class-wide render method thenceforth.

        Any instance without a specific render method set uses the class-wide render
        method.

        NOTE:
            *method* = ``None`` is always allowed, even if the render style doesn't
            implement multiple render methods.
        """
        if method is not None and not isinstance(method, str):
            raise TypeError(
                f"'method' must be a string or `None` (got: {type(method).__name__!r})"
            )

        if method is not None and method.lower() not in self_or_cls._render_methods:
            cls = (
                type(self_or_cls) if isinstance(self_or_cls, __class__) else self_or_cls
            )
            raise ValueError(f"Unknown render method {method!r} for {cls.__name__}")

        if not method:
            if isinstance(self_or_cls, __class__):
                try:
                    del self_or_cls._render_method
                except AttributeError:
                    pass
            elif self_or_cls._render_methods:
                self_or_cls._render_method = self_or_cls._default_render_method
        else:
            self_or_cls._render_method = method

    def set_size(
        self,
        width: Union[int, Size, None] = None,
        height: Union[int, Size, None] = None,
        h_allow: int = 0,
        v_allow: int = 2,
        maxsize: Optional[Tuple[int, int]] = None,
    ) -> None:
        """Sets the image size with extended control.

        Args:
            width: Can be

              * an ``int``; horizontal dimension of the image, in columns.
              * a :py:class:`~term_image.image.Size` enum member.

            height: Can be

              * an ``int``; vertical dimension of the image, in lines.
              * a :py:class:`~term_image.image.Size` enum member.

            h_allow: Horizontal allowance i.e minimum number of columns to leave unused.
            v_allow: Vertical allowance i.e minimum number of lines to leave unused.
            maxsize: If given, as ``(columns, lines)``, it's used instead of the
              terminal size.

        Raises:
            TypeError: An argument is of an inappropriate type.
            ValueError: An argument is of an appropriate type but has an
              unexpected/invalid value.
            ValueError: Both *width* and *height* are specified.
            ValueError: The :term:`available size` is too small for
              :term:`automatic sizing`.
            term_image.exceptions.InvalidSizeError: *maxsize* is given and the
              resulting size will not fit into it.

        If neither *width* nor *height* is given (or both are ``None``),
        :py:attr:`~term_image.image.Size.FIT` applies.

        If *width* or *height* is a :py:class:`~term_image.image.Size` enum
        member, :term:`automatic sizing` applies as prescribed by the enum member.

        When :py:attr:`~term_image.image.Size.FIT_TO_WIDTH` is given,

        * size validation operations take it into consideration.
        * :term:`Vertical allowance` is nullified.

        :term:`Allowances <allowance>` are ignored when *maxsize* is given.

        Image formatting and size validation operations recognize and respect the
        horizontal and vertical allowances, until the image size is re-set.

        NOTE:
           The size is checked to fit in only when *maxsize* is given along with a fixed
           *width* or *height* because :py:meth:`draw` is generally not the means of
           drawing such an image and all rendering methods don't perform any sort of
           size validation.

           If the validation is not desired, specify only one of *maxsize* and
           *width* or *height*, not both.
        """
        if width is not None is not height:
            raise ValueError("Cannot specify both width and height")
        for argname, x in zip(("width", "height"), (width, height)):
            if not (x is None or isinstance(x, (Size, int))):
                raise TypeError(
                    f"{argname!r} must be `None`, a `Size` enum member or an integer "
                    f"(got: type {type(x).__name__!r})"
                )
            if isinstance(x, int) and x <= 0:
                raise ValueError(f"{argname!r} must be positive (got: {x})")

        for argname, x in zip(("h_allow", "v_allow"), (h_allow, v_allow)):
            if not isinstance(x, int):
                raise TypeError(
                    f"{argname!r} must be an integer (got: type {type(x).__name__!r})"
                )
            if x < 0:
                raise ValueError(f"{argname!r} must be non-negative (got: {x})")

        if maxsize is not None:
            if not (
                isinstance(maxsize, tuple) and all(isinstance(x, int) for x in maxsize)
            ):
                raise TypeError(
                    f"'maxsize' must be a tuple of integers (got: {maxsize!r})"
                )

            if not (len(maxsize) == 2 and all(x > 0 for x in maxsize)):
                raise ValueError(
                    f"'maxsize' must contain two positive integers (got: {maxsize})"
                )

        fit_to_width = Size.FIT_TO_WIDTH in (width, height)
        self._size = self._valid_size(
            width,
            height,
            h_allow,
            v_allow * (not fit_to_width),
            maxsize,
        )
        self._fit_to_width = fit_to_width
        self._h_allow = h_allow
        self._v_allow = v_allow * (not fit_to_width)

    def tell(self) -> int:
        """Returns the current image frame number."""
        return self._seek_position if self._is_animated else 0

    # Private Methods

    @classmethod
    def _check_format_spec(
        cls, spec: str
    ) -> Tuple[
        Optional[str],
        Optional[int],
        Optional[str],
        Optional[int],
        Union[None, float, str],
        Dict[str, Any],
    ]:
        """Validates a format specifier and translates it into the required values.

        Returns:
            A tuple ``(h_align, width, v_align, height, alpha, style_args)`` containing
            values as required by ``_format_render()`` and ``_render_image()``.
        """
        match_ = _FORMAT_SPEC.fullmatch(spec)
        if not match_ or _NO_VERTICAL_SPEC.fullmatch(spec):
            raise ValueError(f"Invalid format specifier (got: {spec!r})")

        (
            _,
            h_align,
            width,
            _,
            v_align,
            height,
            alpha,
            threshold_or_bg,
            _,
            style_spec,
        ) = match_.groups()

        return (
            *cls._check_formatting(
                h_align, width and int(width), v_align, height and int(height)
            ),
            (
                threshold_or_bg
                and (
                    "#" + threshold_or_bg.lstrip("#")
                    if _ALPHA_BG_FORMAT.fullmatch("#" + threshold_or_bg.lstrip("#"))
                    else float(threshold_or_bg)
                )
                if alpha
                else _ALPHA_THRESHOLD
            ),
            style_spec and cls._check_style_format_spec(style_spec, style_spec) or {},
        )

    @staticmethod
    def _check_formatting(
        h_align: Optional[str] = None,
        width: Optional[int] = None,
        v_align: Optional[str] = None,
        height: Optional[int] = None,
    ) -> Tuple[Union[None, str, int]]:
        """Validates formatting arguments while also translating literal ones.

        Returns:
            The respective arguments appropriate for ``_format_render()``.
        """
        if not isinstance(h_align, (type(None), str)):
            raise TypeError("'h_align' must be a string.")
        if None is not h_align not in set("<|>"):
            align = {"left": "<", "center": "|", "right": ">"}.get(h_align)
            if not align:
                raise ValueError(f"Invalid horizontal alignment option: {h_align!r}")
            h_align = align

        if not isinstance(width, (type(None), int)):
            raise TypeError("Padding width must be `None` or an integer.")
        if width is not None:
            if width <= 0:
                raise ValueError(f"Padding width must be positive (got: {width})")

        if not isinstance(v_align, (type(None), str)):
            raise TypeError("'v_align' must be a string.")
        if None is not v_align not in set("^-_"):
            align = {"top": "^", "middle": "-", "bottom": "_"}.get(v_align)
            if not align:
                raise ValueError(f"Invalid vertical alignment option: {v_align!r}")
            v_align = align

        if not isinstance(height, (type(None), int)):
            raise TypeError("Padding height must be `None` or an integer.")
        if None is not height <= 0:
            raise ValueError(f"Padding height must be positive (got: {height})")

        return h_align, width, v_align, height

    @staticmethod
    def _check_scale(scale: Tuple[float, float]) -> Tuple[float, float]:
        """Checks a tuple of scale values.

        Returns:
            The tuple of scale values, if valid.

        Raises:
            TypeError: The object is not a tuple of ``float``\\ s.
            ValueError: The object is not a 2-tuple or the values are out of range.
        """
        if not (isinstance(scale, tuple) and all(isinstance(x, float) for x in scale)):
            raise TypeError(f"'scale' must be a tuple of floats (got: {scale!r})")

        if not (len(scale) == 2 and all(0.0 < x <= 1.0 for x in scale)):
            raise ValueError(
                f"'scale' must be a tuple of two floats, 0.0 < x <= 1.0 (got: {scale})"
            )
        return scale

    @staticmethod
    def _check_scale_2(value: float) -> float:
        """Checks a single scale value.

        Returns:
            The scale value, if valid.

        Raises:
            TypeError: The object is not a ``float``.
            ValueError: The value is out of range.
        """
        if not isinstance(value, float):
            raise TypeError(
                f"Given value must be a float (got: type {type(value).__name__!r})"
            )
        if not 0.0 < value <= 1.0:
            raise ValueError(f"Scale value out of range (got: {value})")
        return value

    @classmethod
    def _check_style_args(cls, style_args: Dict[str, Any]) -> Dict[str, Any]:
        """Validates style-specific arguments and translates them into the required
        values.

        Removes any argument having a value equal to the default.

        Returns:
            A mapping of keyword arguments.

        Raises:
            TypeError: An argument is of an inappropriate type.
            ValueError: An argument is of an appropriate type but has an
              unexpected/invalid value.
            term_image.exceptions.StyleError: An unknown style-specific parameter is
              given.
        """
        for name, value in tuple(style_args.items()):
            try:
                (
                    default,
                    (check_type, type_msg),
                    (check_value, value_msg),
                ) = cls._style_args[name]
            except KeyError:
                for other_cls in cls.__mro__:
                    # less costly than memebership tests on every class' __bases__
                    if other_cls is __class__:
                        raise _style_error(cls)(
                            f"Unknown style-specific parameter {name!r}"
                        )

                    if not issubclass(
                        other_cls, __class__
                    ) or "_style_args" not in vars(other_cls):
                        continue

                    try:
                        (check_type, type_msg), (check_value, value_msg) = super(
                            other_cls, cls
                        )._style_args[name]
                        break
                    except KeyError:
                        pass
                else:
                    raise _style_error(cls)(
                        f"Unknown style-specific parameter {name!r}"
                    )

            if not check_type(value):
                raise TypeError(f"{type_msg} (got: {type(value).__name__})")
            if not check_value(value):
                raise ValueError(f"{value_msg} (got: {value!r})")

            # Must not occur before type and value checks to avoid falling prey of
            # operator overloading
            if value == default:
                del style_args[name]

        return style_args

    @classmethod
    def _check_style_format_spec(cls, spec: str, original: str) -> Dict[str, Any]:
        """Validates a style-specific format specifier and translates it into
        the required values.

        Returns:
            A mapping of keyword arguments.

        Raises:
            term_image.exceptions.StyleError: Invalid style-specific format specifier.

        **Every style-specific format spec should be handled as follows:**

        Every overriding method should call the overriden method (more on this below).
        At every step in the call chain, the specifier should be of the form::

            [parent] [current] [invalid]

        where:

        - *parent* is the portion to be interpreted at an higher level in the chain
        - *current* is the portion to be interpreted at the current level in the chain
        - the *invalid* portion determines the validity of the format spec

        Handle the portions in the order *invalid*, *parent*, *current*, so that
        validity can be determined before any further processing.

        At any point in the chain where the *invalid* portion exists (i.e is non-empty),
        the format spec can be correctly taken to be invalid.

        An overriding method must call the overridden method with the *parent* portion
        and the original format spec, **if** *parent* **is not empty**, such that every
        successful check ends up at `BaseImage._check_style_args()` or when *parent* is
        empty.

        :py:meth:`_get_style_format_spec` can (optionally) be used to parse the format
        spec at each level of the call chain.
        """
        if spec:
            raise _style_error(cls)(
                f"Invalid style-specific format specifier {original!r}"
                + (f", detected at {spec!r}" if spec != original else "")
            )
        return {}

    @classmethod
    def _clear_frame(cls) -> bool:
        """Clears an animation frame on-screen.

        Called by :py:meth:`_display_animated` just before drawing a new frame.

        | Only required by styles wherein an image is not overwritten by another image
          e.g some graphics-based styles.
        | The base implementation does nothing and should be overridden only if
          required.

        Returns:
            ``True`` if the frame was cleared. Otherwise, ``False``.
        """
        return False

    def _display_animated(
        self,
        img: PIL.Image.Image,
        alpha: Union[None, float, str],
        fmt: Tuple[Union[None, str, int]],
        repeat: int,
        cached: Union[bool, int],
        **style_args: Any,
    ) -> None:
        """Displays an animated GIF image in the terminal.

        NOTE:
            This is done indefinitely but can be terminated with ``Ctrl-C``
            (``SIGINT``), raising ``KeyboardInterrupt``.
        """
        lines = max(
            (fmt or (None,))[-1] or get_terminal_size()[1] - self._v_allow,
            self.rendered_height,
        )
        prev_seek_pos = self._seek_position
        duration = self._frame_duration
        image_it = ImageIterator(self, repeat, "", cached)
        image_it._animator = image_it._animate(img, alpha, fmt, style_args)

        try:
            print(next(image_it._animator), end="", flush=True)  # First frame

            # Render next frame during current frame's duration
            start = time.time()
            for frame in image_it._animator:  # Renders next frame
                # Left-over of current frame's duration
                time.sleep(max(0, duration - (time.time() - start)))

                # Clear the current frame, if necessary,
                # move cursor up to the begining of the first line of the image
                # and print the new current frame.
                self._clear_frame()
                print(f"\r{CSI}{lines - 1}A", frame, sep="", end="", flush=True)

                # Render next frame during current frame's duration
                start = time.time()
        except (KeyboardInterrupt, Exception):
            self._handle_interrupted_draw()
            raise
        finally:
            if img is not self._source:
                img.close()
            image_it.close()
            self._seek_position = prev_seek_pos
            # Move the cursor to the last line of the image to prevent "overlayed"
            # output in the terminal
            print(f"{CSI}{lines}B", end="")

    def _format_render(
        self,
        render: str,
        h_align: Optional[str] = None,
        width: Optional[int] = None,
        v_align: Optional[str] = None,
        height: Optional[int] = None,
    ) -> str:
        """Formats rendered image text.

        All arguments should be passed through ``_check_formatting()`` first.
        """
        cols, lines = self.rendered_size
        terminal_size = get_terminal_size()

        width = width or terminal_size[0] - self._h_allow
        if width > cols:
            if h_align == "<":  # left
                left = ""
                right = " " * (width - cols)
            elif h_align == ">":  # right
                left = " " * (width - cols)
                right = ""
            else:  # center
                left = " " * ((width - cols) // 2)
                right = " " * (width - cols - len(left))
            render = render.replace("\n", f"{right}\n{left}")
        else:
            left = right = ""

        height = height or terminal_size[1] - self._v_allow
        if height > lines:
            if v_align == "^":  # top
                top = 0
                bottom = height - lines
            elif v_align == "_":  # bottom
                top = height - lines
                bottom = 0
            else:  # middle
                top = (height - lines) // 2
                bottom = height - lines - top
            top = f"{' ' * width}\n" * top
            bottom = f"\n{' ' * width}" * bottom
        else:
            top = bottom = ""

        return (
            "".join((top, left, render, right, bottom))
            if width > cols or height > lines
            else render
        )

    @_close_validated
    def _get_image(self) -> PIL.Image.Image:
        """Returns the PIL image instance corresponding to the image source as-is"""
        return (
            Image.open(self._source) if isinstance(self._source, str) else self._source
        )

    def _get_render_data(
        self,
        img: PIL.Image.Image,
        alpha: Union[None, float, str],
        *,
        size: Optional[Tuple[int, int]] = None,
        pixel_data: bool = True,
        round_alpha: bool = False,
        frame: bool = False,
    ) -> Tuple[
        PIL.Image.Image, Optional[List[Tuple[int, int, int]]], Optional[List[int]]
    ]:
        """Returns the PIL image instance and pixel data required to render an image.

        Args:
            size: If given (in pixels), it is used instead of the pixel-equivalent of
              the image size.
            pixel_data: If ``False``, ``None`` is returned for all pixel data.
            round_alpha: Only applies when *alpha* is a ``float``.

              If ``True``, returned alpha values are bi-level (``0`` or ``255``), based
              on the given alpha threshold.
              Also, the image is blended with the active terminal's BG color (or black,
              if undetermined) while leaving the alpha intact.

            frame: If ``True``, implies *img* is being used by ``ImageIterator``,
              hence, *img* is not closed.

        The returned image is appropriately converted, resized and composited
        (if need be).

        The pixel data are the last two items of the returned tuple ``(rgb, a)``, where:
          * ``rgb`` is a list of ``(r, g, b)`` tuples containing the colour channels of
            the image's pixels in a flattened row-major order where ``r``, ``g``, ``b``
            are integers in the range [0, 255].
          * ``a`` is a list of integers in the range [0, 255] representing the alpha
            channel of the image's pixels in a flattened row-major order.
        """

        def convert_resize_img(mode: str):
            nonlocal img

            if img.mode != mode:
                prev_img = img
                try:
                    img = img.convert(mode)
                except Exception as e:
                    raise ValueError("Unable to convert image") from e
                finally:
                    if frame_img is not prev_img is not self._source:
                        prev_img.close()

            if img.size != size:
                prev_img = img
                try:
                    img = img.resize(size, Image.Resampling.BOX)
                except ValueError as e:
                    raise ValueError("Image size or scale too small") from e
                finally:
                    if frame_img is not prev_img is not self._source:
                        prev_img.close()

        frame_img = img if frame else None
        if self._is_animated:
            img.seek(self._seek_position)
        if not size:
            size = self._get_render_size()

        if alpha is None or img.mode in {"1", "L", "RGB", "HSV", "CMYK"}:
            convert_resize_img("RGB")
            if pixel_data:
                rgb = list(img.getdata())
                a = [255] * mul(*size)
        else:
            convert_resize_img("RGBA")
            if isinstance(alpha, str):
                if alpha == "#":
                    alpha = get_fg_bg_colors(hex=True)[1] or "#000000"
                bg = Image.new("RGBA", img.size, alpha)
                bg.alpha_composite(img)
                if frame_img is not img is not self._source:
                    img.close()
                img = bg.convert("RGB")
                if pixel_data:
                    a = [255] * mul(*size)
            else:
                if pixel_data:
                    a = list(img.getdata(3))
                    if round_alpha:
                        alpha = round(alpha * 255)
                        a = [0 if val < alpha else 255 for val in a]
                if round_alpha:
                    bg = Image.new(
                        "RGBA", img.size, get_fg_bg_colors(hex=True)[1] or "#000000"
                    )
                    bg.alpha_composite(img)
                    bg.putalpha(img.getchannel("A"))
                    if frame_img is not img is not self._source:
                        img.close()
                    img = bg

            if pixel_data:
                rgb = list((img if img.mode == "RGB" else img.convert("RGB")).getdata())

        return (img, *(pixel_data and (rgb, a) or (None, None)))

    @abstractmethod
    def _get_render_size(self) -> Tuple[int, int]:
        """Returns the size (in pixels) required to render the image.

        Applies the image scale.
        """
        raise NotImplementedError

    @staticmethod
    def _handle_interrupted_draw():
        """Performs any neccessary actions when image drawing is interrupted."""

    @classmethod
    def _get_style_format_spec(
        cls, spec: str, original: str
    ) -> Tuple[str, List[Union[None, str, Tuple[Optional[str]]]]]:
        """Parses a style-specific format specifier.

        See :py:meth:`_check_format_spec`.

        Returns:
            The *parent* portion and a list of matches for the respective fields of the
            *current* portion of the spec.

            * Any absent field of *current* is ``None``.
            * For a field containing groups, the match, if present, is a tuple
              containing the full match followed by the matches for each group.
            * All matches are in the same order as the fields (including their groups).

        Raises:
            term_image.exceptions.StyleError: The *invalid* portion exists.

        NOTE:
            Please avoid common fields in the format specs of parent and child classes
            (i.e fields that can match the same portion of a given string) as they
            result in ambiguities.
        """
        patterns = iter(cls._FORMAT_SPEC)
        fields = []
        for pattern in patterns:
            match = pattern.search(spec)
            if match:
                fields.append(
                    (match.group(), *match.groups())
                    if pattern.groups
                    else match.group()
                )
                start = match.start()
                end = match.end()
                break
            else:
                fields.append(
                    (None,) * (pattern.groups + 1) if pattern.groups else None
                )
        else:
            start = end = len(spec)

        for pattern in patterns:
            match = pattern.match(spec, pos=end)
            if match:
                fields.append(
                    (match.group(), *match.groups())
                    if pattern.groups
                    else match.group()
                )
                end = match.end()
            else:
                fields.append(
                    (None,) * (pattern.groups + 1) if pattern.groups else None
                )

        parent, invalid = spec[:start], spec[end:]
        if invalid:
            raise _style_error(cls)(
                f"Invalid style-specific format specifier {original!r}"
                f", detected at {invalid!r}"
            )

        return parent, fields

    @staticmethod
    @abstractmethod
    def _pixels_cols(
        *, pixels: Optional[int] = None, cols: Optional[int] = None
    ) -> int:
        """Returns the number of pixels represented by a given number of columns
        or vice-versa.
        """
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def _pixels_lines(
        *, pixels: Optional[int] = None, lines: Optional[int] = None
    ) -> int:
        """Returns the number of pixels represented by a given number of lines
        or vice-versa.
        """
        raise NotImplementedError

    @abstractmethod
    def _render_image(
        self,
        img: PIL.Image.Image,
        alpha: Union[None, float, str],
        *,
        frame: bool = False,  # For `ImageIterator`
    ) -> str:
        """Converts an image into a string which reproduces the image when printed
        to the terminal.

        NOTE:
            This method is not meant to be used directly, use it via `_renderer()`
            instead.
        """
        raise NotImplementedError

    def _renderer(
        self,
        renderer: FunctionType,
        *args: Any,
        scroll: bool = False,
        check_size: bool = False,
        animated: bool = False,
        **kwargs,
    ) -> Any:
        """Performs common render preparations and a rendering operation.

        Args:
            renderer: The function to perform the specific rendering operation for the
              caller of this method, ``_renderer()``.
              This function must accept at least one positional argument, the
              ``PIL.Image.Image`` instance corresponding to the source.
            args: Positional arguments to pass on to *renderer*, after the
              ``PIL.Image.Image`` instance.
            scroll: See *scroll* in ``draw()``.
            check_size: See *check_size* in ``draw()``.
            animated: If ``True``, *scroll* and *check_size* are ignored and the size
              is validated.
            kwargs: Keyword arguments to pass on to *renderer*.

        Returns:
            The return value of *renderer*.

        Raises:
            term_image.exceptions.InvalidSizeError: *check_size* or *animated* is
              ``True`` and the image's :term:`rendered size` can not fit into the
              :term:`available terminal size <available size>`.
            term_image.exceptions.TermImageError: The image has been finalized.

        NOTE:
            For **non-animations**, if the image size was set with
            :py:attr:term_image.image.Size.FIT_TO_WIDTH`, the image **height** is not
            validated and setting *scroll* is unnecessary.
        """
        _size = self._size
        try:
            if isinstance(_size, Size):
                self.set_size(_size)
            elif check_size or animated:
                # NOTE: If the set size is larger than the available terminal size but
                # the scale makes it fit in, then it's all good.

                columns, lines = map(
                    sub,
                    get_terminal_size(),
                    # *scroll* nullifies vertical allowance for non-animations
                    # Makes a difference when terminal height < vertical allowance
                    (self._h_allow, self._v_allow * (animated or not scroll)),
                )

                if any(
                    map(
                        gt,
                        # the compared height will be 0 if *_fit_to_width* or *scroll*
                        # is `True`. So, the height comparison will always be `False`
                        # since the terminal height should never be < 0.
                        map(
                            mul,
                            self.rendered_size,
                            (1, not (self._fit_to_width or scroll)),
                        ),
                        (columns, lines),
                    )
                ):
                    raise InvalidSizeError(
                        "The "
                        + ("animation" if animated else "image")
                        + " cannot fit into the available terminal size"
                    )

                # Reaching here means it's either valid or *_fit_to_width* and/or
                # *scroll* is/are `True`.
                if animated and self.rendered_height > lines:
                    raise InvalidSizeError(
                        "The rendered height cannot be greater than the terminal "
                        "height for animations"
                    )

            return renderer(self._get_image(), *args, **kwargs)

        finally:
            if isinstance(_size, Size):
                self.size = _size

    def _valid_size(
        self,
        width: Union[int, Size, None] = None,
        height: Union[int, Size, None] = None,
        h_allow: int = 0,
        v_allow: int = 2,
        maxsize: Optional[Tuple[int, int]] = None,
    ) -> Tuple[int, int]:
        """Returns an image size tuple.

        See the description of ``set_size()`` for the parameters.
        """
        ori_width, ori_height = self._original_size
        columns, lines = maxsize or map(sub, get_terminal_size(), (h_allow, v_allow))
        max_width = self._pixels_cols(cols=columns)
        max_height = self._pixels_lines(lines=lines)

        # NOTE: The image scale is not considered since it should never be > 1

        # As for cell ratio...
        #
        # Take for example, pixel ratio = 2.0
        # (i.e cell ratio = 1.0; square character cells).
        # To adjust the image to the proper scale, we either reduce the
        # width (i.e divide by 2.0) or increase the height (i.e multiply by 2.0).
        #
        # On the other hand, if the pixel ratio = 0.5
        # (i.e cell ratio = 0.25; vertically oblong character cells).
        # To adjust the image to the proper scale, we either increase the width
        # (i.e divide by the 0.5) or reduce the height (i.e multiply by the 0.5).
        #
        # Therefore, for the height, we always multiply by the pixel ratio
        # and for the width, we always divide by the pixel ratio.
        # The non-constraining axis is always the one directly adjusted.

        if all(not isinstance(x, int) for x in (width, height)):
            for name in ("columns", "lines"):
                if locals()[name] <= 0:
                    raise ValueError(f"Amount of available {name} too small")

            if Size.AUTO in (width, height):
                width = height = (
                    Size.FIT
                    if (
                        ori_width > max_width
                        or round(ori_height * self._pixel_ratio) > max_height
                    )
                    else Size.ORIGINAL
                )
            elif Size.FIT_TO_WIDTH in (width, height):
                return (
                    self._pixels_cols(pixels=max_width),
                    self._pixels_lines(
                        pixels=round(
                            self._width_height_px(w=max_width) * self._pixel_ratio
                        )
                    ),
                )

            if Size.ORIGINAL in (width, height):
                return (
                    self._pixels_cols(pixels=ori_width),
                    self._pixels_lines(pixels=round(ori_height * self._pixel_ratio)),
                )

            # The smaller fraction will fit on both axis.
            # Hence, the axis with the smaller ratio is the constraining axis.
            # Constraining by the axis with the larger ratio will cause the image
            # to not fit into the axis with the smaller ratio.
            x = max_width / ori_width
            y = max_height / ori_height
            _width_px = ori_width * min(x, y)
            _height_px = ori_height * min(x, y)

            # The cell ratio should affect the axis with the larger ratio since the axis
            # the smaller ratio is already fully occupied

            if x < y:
                _height_px = _height_px * self._pixel_ratio
                # If height becomes greater than the max, reduce it to the max
                height_px = min(_height_px, max_height)
                # Calculate the corresponding width
                width_px = round((height_px / _height_px) * _width_px)
                # Round the height
                height_px = round(height_px)
            else:
                _width_px = _width_px / self._pixel_ratio
                # If width becomes greater than the max, reduce it to the max
                width_px = min(_width_px, max_width)
                # Calculate the corresponding height
                height_px = round((width_px / _width_px) * _height_px)
                # Round the width
                width_px = round(width_px)
            return (
                self._pixels_cols(pixels=width_px),
                self._pixels_lines(pixels=height_px),
            )
        elif width is None:
            width_px = round(
                self._width_height_px(h=self._pixels_lines(lines=height))
                / self._pixel_ratio
            )
            width = self._pixels_cols(pixels=width_px)
        elif height is None:
            height_px = round(
                self._width_height_px(w=self._pixels_cols(cols=width))
                * self._pixel_ratio
            )
            height = self._pixels_lines(pixels=height_px)

        if maxsize and (width > columns or height > lines):
            raise InvalidSizeError(
                f"The resulting size {width, height} will not fit into "
                f"'maxsize' {maxsize}"
            )

        return (width, height)

    def _width_height_px(
        self, *, w: Optional[int] = None, h: Optional[int] = None
    ) -> float:
        """Converts the given width (in pixels) to the **unrounded** proportional height
        (in pixels) OR vice-versa.
        """
        ori_width, ori_height = self._original_size
        return (
            (w / ori_width) * ori_height
            if w is not None
            else (h / ori_height) * ori_width
        )


class GraphicsImage(BaseImage):
    """Base of all render styles using terminal graphics protocols.

    Raises:
        term_image.exceptions.StyleError: The :term:`active terminal` doesn't support
          the render style.

    See :py:class:`BaseImage` for the description of the constructor.

    ATTENTION:
        This class cannot be directly instantiated. Image instances should be created
        from its subclasses.
    """

    # Size unit conversion already involves cell size calculation
    _pixel_ratio: float = 1.0

    def __init__(self, image: PIL.Image.Image, **kwargs) -> None:
        if not self.is_supported():
            raise _style_error(type(self))(
                "This image render style is not supported in the active terminal"
            )
        super().__init__(image, **kwargs)

    def _get_minimal_render_size(self, *, adjust: bool) -> Tuple[int, int]:
        render_size = self._get_render_size()
        r_height = self.rendered_height
        width, height = (
            render_size
            if mul(*render_size) < mul(*self._original_size)
            else self._original_size
        )

        # When `_original_size` is used, ensure the height is a multiple of the rendered
        # height, so that pixels can be evenly distributed among all lines.
        # If r_height == 0, height == 0, extra == 0; Handled in `_get_render_data()`.
        if adjust:
            extra = height % (r_height or 1)
            if extra:
                # Incremented to the greater multiple to avoid losing any data
                height = height - extra + r_height

        return width, height

    def _get_render_size(self) -> Tuple[int, int]:
        return tuple(map(mul, self.rendered_size, get_cell_size() or (1, 2)))

    @staticmethod
    def _pixels_cols(
        *, pixels: Optional[int] = None, cols: Optional[int] = None
    ) -> int:
        return (
            ceil(pixels // (get_cell_size() or (1, 2))[0])
            if pixels is not None
            else cols * (get_cell_size() or (1, 2))[0]
        )

    @staticmethod
    def _pixels_lines(
        *, pixels: Optional[int] = None, lines: Optional[int] = None
    ) -> int:
        return (
            ceil(pixels // (get_cell_size() or (1, 2))[1])
            if pixels is not None
            else lines * (get_cell_size() or (1, 2))[1]
        )


class TextImage(BaseImage):
    """Base of all render styles using ASCII/Unicode symbols [with ANSI color codes].

    See :py:class:`BaseImage` for the description of the constructor.

    IMPORTANT:
        Instantiation of subclasses is always allowed, even if the current terminal
        does not [fully] support the render style.

        To check if the render style is fully supported in the current terminal, use
        :py:meth:`is_supported() <BaseImage.is_supported>`.

    ATTENTION:
        This class cannot be directly instantiated. Image instances should be created
        from its subclasses.
    """

    # Pixels are represented in a 1-to-2 ratio within one character cell
    # pixel-size == width * height/2
    # pixel-ratio == width / (height/2) == 2 * (width / height) == 2 * cell-ratio
    _pixel_ratio = property(lambda _: get_cell_ratio() * 2)

    @staticmethod
    @cached
    def _is_on_kitty() -> bool:
        return get_terminal_name_version()[0] == "kitty"


class ImageIterator:
    """Effeciently iterate over :term:`rendered` frames of an :term:`animated` image

    Args:
        image: Animated image.
        repeat: The number of times to go over the entire image. A negative value
          implies infinite repetition.
        format: The :ref:`format specifier <format-spec>` to be used to format the
          rendered frames (default: auto).
        cached: Determines if the :term:`rendered` frames will be cached (for speed up
          of subsequent renders) or not.

          * If ``bool``, it directly sets if the frames will be cached or not.
          * If ``int``, caching is enabled only if the framecount of the image
            is less than or equal to the given number.

    Raises:
        TypeError: An argument is of an inappropriate type.
        ValueError: An argument is of an appropriate type but has an
          unexpected/invalid value.
        term_image.exceptions.StyleError: Invalid style-specific format specifier.

    * If *repeat* equals ``1``, caching is disabled.
    * The iterator has immediate response to changes in the image size
      and :term:`scale`.
    * If the image size is :term:`dynamic <dynamic size>`, it's computed per frame.
    * The number of the last yielded frame is set as the image's seek position.
    * Directly adjusting the seek position of the image doesn't affect iteration.
      Use :py:meth:`ImageIterator.seek` instead.
    * After the iterator is exhausted, the underlying image is set to frame ``0``.
    """

    def __init__(
        self,
        image: BaseImage,
        repeat: int = -1,
        format: str = "",
        cached: Union[bool, int] = 100,
    ) -> None:
        if not isinstance(image, BaseImage):
            raise TypeError(f"Invalid type for 'image' (got: {type(image).__name__})")
        if not image._is_animated:
            raise ValueError("This image is not animated")

        if not isinstance(repeat, int):
            raise TypeError(f"Invalid type for 'repeat' (got: {type(repeat).__name__})")
        if not repeat:
            raise ValueError("'repeat' must be non-zero")

        if not isinstance(format, str):
            raise TypeError(
                "Invalid type for 'format' " f"(got: {type(format).__name__})"
            )
        *fmt, alpha, style_args = image._check_format_spec(format)

        if not isinstance(cached, int):  # `bool` is a subclass of `int`
            raise TypeError(f"Invalid type for 'cached' (got: {type(cached).__name__})")
        if False is not cached <= 0:
            raise ValueError("'cached' must be a boolean or a positive integer")

        self._image = image
        self._repeat = repeat
        self._format = format
        self._cached = (
            cached if isinstance(cached, bool) else image.n_frames <= cached
        ) and repeat != 1
        self._loop_no = None
        self._animator = image._renderer(
            self._animate, alpha, fmt, style_args, check_size=False
        )

    def __del__(self) -> None:
        self.close()

    def __iter__(self) -> None:
        return self

    def __next__(self) -> None:
        try:
            return next(self._animator)
        except StopIteration:
            self.close()
            raise StopIteration(
                "Iteration has reached the given repeat count"
            ) from None
        except AttributeError as e:
            if str(e).endswith("'_animator'"):
                raise StopIteration("Iterator exhausted or closed") from None
            else:
                self.close()
                raise
        except Exception:
            self.close()
            raise

    def __repr__(self) -> None:
        return "{}(image={!r}, repeat={}, format={!r}, cached={}, loop_no={})".format(
            type(self).__name__,
            *self.__dict__.values(),
        )

    loop_no = property(
        lambda self: self._loop_no,
        doc="""Iteration repeat countdown

        Changes on the first iteration of each loop, except for infinite iteration
        where it's always ``-1``.
        """,
    )

    def close(self) -> None:
        """Closes the iterator and releases resources used.

        Does not reset the frame number of the underlying image.

        NOTE:
            This method is automatically called when the iterator is exhausted or
            garbage-collected.
        """
        try:
            self._animator.close()
            del self._animator
            if self._img is not self._image._source:
                self._img.close()
            del self._img
        except AttributeError:
            pass

    def seek(self, pos: int) -> None:
        """Sets the frame number to be yielded on the next iteration without affecting
        the repeat count.

        Args:
            pos: Next frame number.

        Raises:
            TypeError: An argument is of an inappropriate type.
            ValueError: An argument is of an appropriate type but has an
              unexpected/invalid value.
            term_image.exceptions.TermImageError: Iteration has not yet started or the
              iterator is exhausted/closed.

        Frame numbers start from ``0`` (zero).
        """
        if not isinstance(pos, int):
            raise TypeError(f"Invalid frame number type (got: {type(pos).__name__})")
        if not 0 <= pos < self._image.n_frames:
            raise ValueError(
                "Frame number out of range "
                f"(got: {pos}, n_frames={self._image._n_frames})"
            )

        try:
            self._animator.send(pos)
        except TypeError:
            raise TermImageError("Iteration has not yet started") from None
        except AttributeError:
            raise TermImageError("Iterator exhausted or closed") from None

    def _animate(
        self,
        img: PIL.Image.Image,
        alpha: Union[None, float, str],
        fmt: Tuple[Union[None, str, int]],
        style_args: Dict[str, Any],
    ) -> None:
        """Returns a generator that yields rendered and formatted frames of the
        underlying image.
        """
        self._img = img  # For cleanup
        image = self._image
        cached = self._cached
        self._loop_no = repeat = self._repeat
        if cached:
            cache = [(None,) * 2] * image.n_frames

        sent = None
        n = 0
        while repeat:
            if sent is None:
                image._seek_position = n
                try:
                    frame = image._format_render(
                        image._render_image(img, alpha, frame=True, **style_args), *fmt
                    )
                except EOFError:
                    image._seek_position = n = 0
                    if repeat > 0:  # Avoid infinitely large negative numbers
                        self._loop_no = repeat = repeat - 1
                    if cached:
                        break
                    continue
                else:
                    if cached:
                        cache[n] = (frame, hash(image.rendered_size))

            sent = yield frame
            n = n + 1 if sent is None else sent - 1

        if cached:
            n_frames = len(cache)
        while repeat:
            while n < n_frames:
                if sent is None:
                    image._seek_position = n
                    frame, size_hash = cache[n]
                    if hash(image.rendered_size) != size_hash:
                        frame = image._format_render(
                            image._render_image(img, alpha, frame=True, **style_args),
                            *fmt,
                        )
                        cache[n] = (frame, hash(image.rendered_size))

                sent = yield frame
                n = n + 1 if sent is None else sent - 1

            image._seek_position = n = 0
            if repeat > 0:  # Avoid infinitely large negative numbers
                self._loop_no = repeat = repeat - 1

        # For consistency in behaviour
        if img is image._source:
            img.seek(0)



"""Support for command-line execution using `python -m term-image`"""

from __future__ import annotations

import logging as _logging
import multiprocessing
import sys
from threading import Event

from .exit_codes import FAILURE, INTERRUPTED, codes
from .utils import write_tty


def main() -> int:
    """CLI execution entry-point"""
    from . import cli, logging, notify
    from .tui import main

    def finish_loading():
        if not logging.QUIET and notify.loading_indicator:
            notify.end_loading()
            if not main.loop:  # TUI was not launched
                while notify.is_loading():
                    pass
                notify.end_loading()
            notify.loading_indicator.join()

    def finish_multi_logging():
        if logging.MULTI:
            from .logging_multi import child_processes, log_queue

            if log_queue:  # Multi-logging has been successfully initialized
                for process in child_processes:
                    process.join()
                log_queue.put((None,) * 2)  # End of logs
                log_queue.join()

    # 1. `PIL.Image.open()` seems to cause forked child processes to block when called
    # in both the parent and the child.
    # 2. Unifies things across multiple platforms.
    multiprocessing.set_start_method("spawn")

    # Can't use "term_image", since the logger's level is changed.
    # Otherwise, it would affect children of "term_image".
    logger = _logging.getLogger("term-image")
    logger.setLevel(_logging.INFO)

    cli.interrupted = Event()
    try:
        write_tty(b"\033[22;2t")  # Save window title
        write_tty(b"\033]2;Term-Image\033\\")  # Set window title
        exit_code = cli.main()
    except KeyboardInterrupt:
        cli.interrupted.set()  # Signal interruption to subprocesses and other threads.
        finish_loading()
        finish_multi_logging()
        logging.log(
            "Session interrupted",
            logger,
            _logging.CRITICAL,
            # If logging has been successfully initialized
            file=logging.VERBOSE is not None,
            # If the TUI was not launched, only print to console if verbosity is enabled
            direct=bool(main.loop or cli.args and (cli.args.verbose or cli.args.debug)),
        )
        if cli.args and cli.args.debug:
            raise
        return INTERRUPTED
    except Exception as e:
        cli.interrupted.set()  # Signal interruption to subprocesses and other threads.
        finish_loading()
        finish_multi_logging()
        if logging.VERBOSE is not None:  # logging has been successfully initialized
            logger.exception("Session terminated due to:")
        logging.log(
            "Session not ended successfully: "
            f"({type(e).__module__}.{type(e).__qualname__}) {e}",
            logger,
            _logging.CRITICAL,
            # If logging has been successfully initialized
            file=logging.VERBOSE is not None,
        )
        if cli.args and cli.args.debug:
            raise
        return FAILURE
    else:
        finish_loading()
        finish_multi_logging()
        logger.info(f"Session ended with return-code {exit_code} ({codes[exit_code]})")
        return exit_code
    finally:
        write_tty(b"\033[22;2t")  # Restore window title
        # Explicit cleanup is neccessary since the top-level `Image` widgets
        # will still hold references to the `BaseImage` instances
        if cli.url_images:
            for _, value in cli.url_images:
                value._ti_image.close()


if __name__ == "__main__":
    sys.exit(main())



"""CLI argument parsers"""

import argparse
import sys

from . import __version__
from .config import config_options
from .image import ITerm2Image, Size
from .image.common import _ALPHA_THRESHOLD

parser = argparse.ArgumentParser(
    prog="term-image",
    formatter_class=argparse.RawDescriptionHelpFormatter,
    description="Display/Browse images in a terminal",
    epilog=""" \

'--' should be used to separate positional arguments that begin with an '-' \
from options/flags, to avoid ambiguity.
For example, `$ term-image [options] -- -image.jpg --image.png`

Render Styles:
  auto: The best style is automatically determined based on the detected terminal
      support.
  kitty: Uses the kitty graphics protocol. Currently supported terminal emulators
      include (but might not be limited to):
      - Kitty >= 0.20.0
      - Konsole >= 22.04.0
  iterm2: Uses the iTerm2 inline image protocol. Currently supported terminal
      emulators include (but might not be limited to):
      - iTerm2
      - Konsole >= 22.04.0
      - WezTerm
  block: Uses unicode half blocks with 24-bit color escape codes to represent images
      with a density of two pixels per character cell.

  Using a terminal graphics-based style not supported by the active terminal is not
  allowed by default. To force a style that is normally unsupported, add the
  '--force-style' flag.

FOOTNOTES:
  1. Width and height are in units of columns and lines repectively.
     The "available terminal size" is the current terminal size minus allowances.
     By default (i.e if none of the sizing options is specified), the equivalent of
     `--original-size` is used if not larger than the available size, else `--fit`.
  2. The size is multiplied by the scale on respective axes when an image is rendered.
     A scale value must be such that 0.0 < value <= 1.0.
  3. In CLI mode, only image sources are used, directory sources are skipped.
     Animated images are displayed only when animation is disabled (with `--no-anim`),
     when there's only one image source or when using native animation of some render
     styles.
  4. Any image having more pixels than the specified maximum will be:
     - skipped, in CLI mode, if '--max-pixels-cli' is specified.
     - replaced, in TUI mode, with a placeholder when displayed but can still be forced
     to display or viewed externally.
     Note that increasing this should not have any effect on general performance
     (i.e navigation, etc) but the larger an image is, the more the time and memory
     it'll take to render it. Thus, a large image might delay the rendering of other
     images to be rendered immediately after it.
  5. Frames will not be cached for any animation with more frames than this value.
     Memory usage depends on the frame count per image, not this maximum count.
  6. 0 -> worst quality; smallest data size, 95 -> best quality; largest data size.
     Reduces render time & image data size and increases drawing speed on the terminal's
     end but at the cost of image quality and color reproduction. Useful for animations
     with high pixel density / color sparseness.
     This option only applies when an image is re-encoded and not read directly from
     file (see `--iterm2-no-read-from-file`). By default (i.e when disabled), PNG format
     is used for re-encoding images, which has less compression with better quality.
     JPEG format can only be used for non-transparent images but the transparency status
     of some images can not be correctly determined in an efficient way at render time.
     Thus, to ensure the JPEG format is always used for re-encoded images, disable
     transparency (`--no-alpha`) or set a background color (`-b/--alpha-bg`).
  7. By default, image data is used directly from file when no image manipulation is
     required. Otherwise, it's re-encoded in PNG (or JPEG, if enabled) format.
     Significantly reduces render time when applicable. This option does not apply to
     animations, native or not.
  8. Any event with a level lower than the specified one is not reported.
  9. Supports all image formats supported by `PIL.Image.open()`.
     See https://pillow.readthedocs.io/en/latest/handbook/image-file-formats.html for
     details.
""",
    add_help=False,  # '-h' is used for HEIGHT
)

# General
general = parser.add_argument_group("General Options")
general.add_argument(
    "--help",
    action="help",
    help="Show this help message and exit",
)
general.add_argument(
    "--version",
    action="version",
    version=__version__,
    help="Show the program version and exit",
)
general.add_argument(
    "--query-timeout",
    type=float,
    metavar="N",
    help=(
        "Timeout (in seconds) for all terminal queries "
        f"(default: {config_options.query_timeout})"
    ),
)
general.add_argument(
    "-S",
    "--style",
    choices=("auto", "block", "iterm2", "kitty"),
    help=(
        f"Image render style (default: {config_options.style}). "
        'See "Render Styles" below'
    ),
)
general.add_argument(
    "--force-style",
    action="store_true",
    help=(
        "Use the specified render style even if it's reported as unsupported by "
        "the active terminal"
    ),
)

cell_ratio_options = general.add_mutually_exclusive_group()
cell_ratio_options.add_argument(
    "-C",
    "--cell-ratio",
    type=float,
    metavar="N",
    help=(
        "The width-to-height ratio of a character cell in the terminal, to "
        "preserve image aspect ratio (default: auto)"
    ),
)
cell_ratio_options.add_argument(
    "--auto-cell-ratio",
    action="store_true",
    help="Determine the cell ratio from the terminal, if possible (default)",
)

win_size_options = general.add_mutually_exclusive_group()
win_size_options.add_argument(
    "--swap-win-size",
    action="store_true",
    default=None,
    help=(
        "A workaround for 'auto cell ratio' on some terminal emulators (e.g older "
        "VTE-based ones) that wrongly report window dimensions swapped"
    ),
)
win_size_options.add_argument(
    "--no-swap-win-size",
    action="store_false",
    default=None,
    dest="swap_win_size",
    help="Unlike '--swap-win-size', use the reported window size as-is (default)",
)

mode_options = general.add_mutually_exclusive_group()
mode_options.add_argument(
    "--cli",
    action="store_true",
    help=(
        "Do not the launch the TUI, instead draw all image sources "
        "to the terminal directly [3]"
    ),
)
mode_options.add_argument(
    "--tui",
    action="store_true",
    help="Always launch the TUI, even for a single image",
)

# # Animation
anim_options = parser.add_argument_group("Animation Options (General)")
anim_options.add_argument(
    "-f",
    "--frame-duration",
    type=float,
    metavar="N",
    help=(
        "The time (in seconds) between frames for all animated images "
        "(default: determined per image from the metadata OR 0.1)"
    ),
)
anim_options.add_argument(
    "-R",
    "--repeat",
    type=int,
    default=-1,
    metavar="N",
    help=(
        "Number of times to repeat all frames of an animated image; A negative "
        "count implies an infinite loop (default: -1)"
    ),
)

anim_cache_options = anim_options.add_mutually_exclusive_group()
anim_cache_options.add_argument(
    "--anim-cache",
    type=int,
    metavar="N",
    help=(
        "Maximum frame count for animation frames to be cached (Better performance "
        f"at the cost of memory) (default: {config_options.anim_cache}) [5]"
    ),
)
anim_cache_options.add_argument(
    "--cache-all-anim",
    action="store_true",
    help=(
        "Cache frames for all animations (Beware, uses up a lot of memory for "
        "animated images with very high frame count)"
    ),
)
anim_cache_options.add_argument(
    "--cache-no-anim",
    action="store_true",
    help="Disable frame caching (Less memory usage but reduces performance)",
)

anim_options.add_argument(
    "--no-anim",
    action="store_true",
    help=(
        "Disable image animation. Animated images are displayed as just their "
        "first frame."
    ),
)

# # Transparency
_alpha_options = parser.add_argument_group(
    "Transparency Options (General)",
    "NOTE: These are mutually exclusive",
)
alpha_options = _alpha_options.add_mutually_exclusive_group()
alpha_options.add_argument(
    "--no-alpha",
    action="store_true",
    help="Disable image transparency (alpha channel is removed)",
)
alpha_options.add_argument(
    "-A",
    "--alpha",
    type=float,
    metavar="N",
    default=_ALPHA_THRESHOLD,
    help=(
        "Alpha ratio above which pixels are taken as opaque (0 <= x < 1), "
        f"for text-based render styles (default: {_ALPHA_THRESHOLD:f})"
    ),
)
alpha_options.add_argument(
    "-b",
    "--alpha-bg",
    nargs="?",
    const="",
    metavar="COLOR",
    help=(
        "Hex color (without '#') to replace transparent backgrounds with "
        "(omit `COLOR` to use the terminal's default BG color)"
    ),
)

# CLI-only
cli_options = parser.add_argument_group(
    "CLI-only Options",
    "These options apply only when there is only one valid image source or `--cli` "
    "is specified",
)
cli_options.add_argument(
    "--h-allow",
    type=int,
    default=0,
    metavar="N",
    help=(
        "Horizontal allowance i.e minimum number of columns to leave unused "
        "(default: 0)"
    ),
)
cli_options.add_argument(
    "--v-allow",
    type=int,
    default=2,
    metavar="N",
    help=(
        "Vertical allowance i.e minimum number of lines to leave unused " "(default: 2)"
    ),
)
cli_options.add_argument(
    "--scroll",
    action="store_true",
    help=(
        "Allow an image's height to be greater than the terminal height. "
        "Not needed when `--fit-to-width` is specified."
    ),
)
cli_options.add_argument(
    "-O",
    "--oversize",
    action="store_true",
    help=(
        "Allow an image's size to be greater than the terminal size "
        "(To be used with `-w`, `-h` or `--original-size`)"
    ),
)
cli_options.add_argument(
    "-s",
    "--scale",
    type=float,
    metavar="N",
    help="Image scale (overrides `-x` and `-y`) [2]",
)
cli_options.add_argument(
    "-x",
    "--scale-x",
    type=float,
    metavar="N",
    default=1.0,
    help="Image x-axis scale (default: 1.0) [2]",
)
cli_options.add_argument(
    "-y",
    "--scale-y",
    type=float,
    metavar="N",
    default=1.0,
    help="Image y-axis scale (default: 1.0) [2]",
)
cli_options.add_argument(
    "--max-pixels-cli",
    action="store_true",
    help=("Apply '--max-pixels' in CLI mode"),
)

# Sizing
_size_options = parser.add_argument_group(
    "Sizing Options (CLI-only)",
    "These apply to all images and are mutually exclusive [1]",
)
size_options = _size_options.add_mutually_exclusive_group()
size_options.add_argument(
    "-w",
    "--width",
    type=int,
    metavar="N",
    help="Image width",
)
size_options.add_argument(
    "-h",
    "--height",
    type=int,
    metavar="N",
    help="Image height",
)
size_options.add_argument(
    "--fit",
    action="store_const",
    const=Size.FIT,
    dest="auto_size",
    help="Fit each image optimally within the available terminal size",
)
size_options.add_argument(
    "--fit-to-width",
    action="store_const",
    const=Size.FIT_TO_WIDTH,
    dest="auto_size",
    help=(
        "Fit each image to the available terminal width, "
        "`--v-allow` has no effect i.e vertical allowance is ignored"
    ),
)
size_options.add_argument(
    "--original-size",
    action="store_const",
    const=Size.ORIGINAL,
    dest="auto_size",
    help=(
        "Render each image using its original size (See `--oversize`, "
        "USE WITH CAUTION!)"
    ),
)

# # Alignment
align_options = parser.add_argument_group(
    "Alignment Options (CLI-only)",
    "These apply to all images",
)
align_options.add_argument(
    "--no-align",
    action="store_true",
    help=(
        "Output image without alignment or padding. "
        "Overrides all other alignment options"
    ),
)
align_options.add_argument(
    "-H",
    "--h-align",
    choices=("left", "center", "right"),
    help="Horizontal alignment (default: center)",
)
align_options.add_argument(
    "--pad-width",
    metavar="N",
    type=int,
    help=(
        "Number of columns within which to align each image "
        "(default: terminal width, minus horizontal allowance)"
    ),
)
align_options.add_argument(
    "-V",
    "--v-align",
    choices=("top", "middle", "bottom"),
    help="Vertical alignment (default: middle)",
)
align_options.add_argument(
    "--pad-height",
    metavar="N",
    type=int,
    help="Number of lines within which to align each image (default: none)",
)

# TUI-only
tui_options = parser.add_argument_group(
    "TUI-only Options",
    "These options apply only when there is at least one valid directory source, "
    "multiple valid sources or `--tui` is specified",
)

tui_options.add_argument(
    "-a",
    "--all",
    action="store_true",
    help="Inlcude hidden file and directories",
)
tui_options.add_argument(
    "-r",
    "--recursive",
    action="store_true",
    help="Scan for local images recursively",
)
tui_options.add_argument(
    "-d",
    "--max-depth",
    type=int,
    metavar="N",
    default=sys.getrecursionlimit() - 50,
    help=f"Maximum recursion depth (default: {sys.getrecursionlimit() - 50})",
)

# Performance
perf_options = parser.add_argument_group("Performance Options (General)")
perf_options.add_argument(
    "--max-pixels",
    type=int,
    metavar="N",
    help=(
        "Maximum amount of pixels in images to be displayed "
        f"(default: {config_options.max_pixels}) [4]"
    ),
)

perf_options.add_argument(
    "--checkers",
    type=int,
    metavar="N",
    help=(
        "Maximum number of sub-processes for checking directory sources (default: auto)"
    ),
)
perf_options.add_argument(
    "--getters",
    type=int,
    metavar="N",
    help=(
        "Number of threads for downloading images from URL sources "
        f"(default: {config_options.getters})"
    ),
)
perf_options.add_argument(
    "--grid-renderers",
    type=int,
    metavar="N",
    help=(
        "Number of subprocesses for rendering grid cells "
        f"(default: {config_options.grid_renderers})"
    ),
)

multi_options = perf_options.add_mutually_exclusive_group()
multi_options.add_argument(
    "--multi",
    action="store_true",
    default=None,
    help="Enable multiprocessing, if supported (default)",
)
multi_options.add_argument(
    "--no-multi",
    action="store_false",
    default=None,
    dest="multi",
    help="Disable multiprocessing",
)

# Config
config_options__ = parser.add_argument_group(
    "Config Options",
    "NOTE: These are mutually exclusive",
)
config_options_ = config_options__.add_mutually_exclusive_group()

config_options_.add_argument(
    "--config",
    metavar="FILE",
    help="The config file to use for this session (default: Searches XDG Base Dirs)",
)
config_options_.add_argument(
    "--no-config",
    action="store_true",
    help="Use the default configuration",
)

# Logging
log_options_ = parser.add_argument_group(
    "Logging Options",
    "NOTE: All these, except '-l/--log-file', are mutually exclusive",
)
log_options = log_options_.add_mutually_exclusive_group()

log_options_.add_argument(
    "-l",
    "--log-file",
    metavar="FILE",
    help=f"The file to write logs to (default: {config_options.log_file})",
)
log_options.add_argument(
    "--log-level",
    choices=("DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"),
    default="WARNING",
    help="Logging level for the session (default: WARNING) [6]",
)
log_options.add_argument(
    "-q",
    "--quiet",
    action="store_true",
    help="No notifications, except fatal errors",
)
log_options.add_argument(
    "-v",
    "--verbose",
    action="store_true",
    help="More detailed event reporting. Also sets logging level to INFO",
)
log_options.add_argument(
    "--verbose-log",
    action="store_true",
    help="Like --verbose but only applies to the log file",
)
log_options.add_argument(
    "--debug",
    action="store_true",
    help="Implies --log-level=DEBUG with verbosity",
)

# Positional
parser.add_argument(
    "sources",
    nargs="*",
    metavar="source",
    help=(
        "Path(s) to local image(s) and/or directory(s) OR URLs. "
        "If no source is given, the current working directory is used."
    ),
)

kitty_parser = argparse.ArgumentParser(add_help=False)
kitty_options = kitty_parser.add_argument_group(
    "Kitty Style Options",
    "These options apply only when the 'kitty' render style is used",
)
kitty_options.add_argument(
    "--kz",
    "--kitty-z-index",
    metavar="N",
    dest="z_index",
    default=0,
    type=int,
    help=(
        "Image stacking order [CLI-only]; `>= 0` -> above text, `< 0` -> below "
        "text, `< -(2**31)/2` -> below cells with non-default background "
        "(default: 0)"
    ),
)
kitty_options.add_argument(
    "--kc",
    "--kitty-compress",
    metavar="N",
    dest="compress",
    default=4,
    type=int,
    help=(
        "ZLIB compression level; 0 -> no compression, 1 -> best speed, "
        "9 -> best compression (default: 4)"
    ),
)

iterm2_parser = argparse.ArgumentParser(add_help=False)
iterm2_options = iterm2_parser.add_argument_group(
    "iTerm2 Style Options",
    "These options apply only when the 'iterm2' render style is used",
)
iterm2_options.add_argument(
    "--itn",
    "--iterm2-native",
    action="store_true",
    dest="native",
    help="Use iTerm2's native animation; Animations will not be skipped [CLI-only]",
)
iterm2_options.add_argument(
    "--itnm",
    "--iterm2-native-maxsize",
    metavar="N",
    dest="native_maxsize",
    default=ITerm2Image.NATIVE_ANIM_MAXSIZE,
    type=int,
    help=(
        "Maximum size (in bytes) of image data for native animation [CLI-only] "
        f"(default: {ITerm2Image.NATIVE_ANIM_MAXSIZE})"
    ),
)
iterm2_options.add_argument(
    "--itc",
    "--iterm2-compress",
    metavar="N",
    dest="compress",
    default=4,
    type=int,
    help=(
        "ZLIB compression level, for images re-encoded in PNG format "
        "0 -> no compression, 1 -> best speed, 9 -> best compression (default: 4)"
    ),
)
iterm2_options.add_argument(
    "--itjq",
    "--iterm2-jpeg-quality",
    metavar="N",
    dest="jpeg_quality",
    default=ITerm2Image.JPEG_QUALITY,
    type=int,
    help=(
        "JPEG compression status and quality; `< 0` -> disabled, `0 to 95` -> "
        f"quality (default: {ITerm2Image.JPEG_QUALITY}) [6]"
    ),
)
iterm2_options.add_argument(
    "--itnrff",
    "--iterm2-no-read-from-file",
    action="store_false",
    dest="read_from_file",
    help="Never use image data directly from file; always re-encode images [7]",
)

style_parsers = {"kitty": kitty_parser, "iterm2": iterm2_parser}

for style_parser in style_parsers.values():
    parser._actions.extend(style_parser._actions)
    parser._option_string_actions.update(style_parser._option_string_actions)
    parser._action_groups.extend(style_parser._action_groups)
    parser._mutually_exclusive_groups.extend(style_parser._mutually_exclusive_groups)



from setuptools import setup

classifiers = [
    "Environment :: Console",
    "License :: OSI Approved :: MIT License",
    "Intended Audience :: Developers",
    "Intended Audience :: End Users/Desktop",
    "Operating System :: POSIX :: Linux",
    "Operating System :: MacOS",
    "Operating System :: Android",
    "Operating System :: Microsoft :: Windows :: Windows 10",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
    "Topic :: Software Development :: Libraries",
    "Topic :: Utilities",
    "Topic :: Multimedia :: Graphics :: Viewers",
]

with open("README.md", "r") as fp:
    long_description = fp.read()

setup(
    name="term-image",
    version="0.5.0",
    author="AnonymouX47",
    author_email="anonymoux47@gmail.com",
    url="https://github.com/AnonymouX47/term-image",
    description="Display images in the terminal",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    classifiers=classifiers,
    python_requires=">=3.7",
    install_requires=["pillow>=9.1,<10.0", "requests>=2.23,<3.0", "urwid>=2.1,<3.0"],
    entry_points={
        "console_scripts": ["term-image=term_image.__main__:main"],
    },
    project_urls={
        "Changelog": "https://github.com/AnonymouX47/term-image/blob/main/CHANGELOG.md",
        "Documentation": "https://term-image.readthedocs.io/",
        "Funding": "https://github.com/AnonymouX47/term-image#donate",
        "Source": "https://github.com/AnonymouX47/term-image",
        "Tracker": "https://github.com/AnonymouX47/term-image/issues",
    },
    keywords=[
        "image",
        "terminal",
        "viewer",
        "PIL",
        "Pillow",
        "console",
        "xterm",
        "library",
        "cli",
        "tui",
        "ANSI",
        "ASCII-Art",
        "kitty",
        "iterm2",
        "sixel",
        "graphics",
    ],
)



"""Exit codes for fatal errors"""

from __future__ import annotations

codes = {
    0: "SUCCESS",
    1: "FAILURE",
    2: "INVALID_ARG",
    3: "INTERRUPTED",
    4: "NO_VALID_SOURCE",
}

globals().update((v, k) for k, v in codes.items())



"""Term-Image's Configuration"""

from __future__ import annotations

import json
import logging as _logging
import os
from copy import deepcopy
from dataclasses import dataclass, field
from os import path
from typing import Any, Callable, Dict, Optional, Tuple

import urwid

from . import logging, notify
from .utils import QUERY_TIMEOUT, is_writable


class ConfigOptions(dict):
    """Config options store

    * Subscription with an option name returns the corresponding :py:class:`Option`
      instance.
    * Attribute reference with a variable name ('s/ /_/g') returns the option's current
      value.
    * Attribute reference with a "private" name ('s/ /_/g' and preceded by '_') returns
      the option's default value.
    """

    def _attr_to_option(self, attr: str) -> Tuple[Option, str]:
        default = attr.startswith("_")
        name = attr.replace("_", " ")
        if default:
            name = name[1:]
        try:
            return self[name], "default" if default else "value"
        except KeyError:
            raise AttributeError(f"Ain't no such config option as {name!r}") from None

    def __getattr__(self, attr: str):
        return getattr(*self._attr_to_option(attr))

    def __setattr__(self, attr: str, value: Any):
        setattr(*self._attr_to_option(attr), value)


@dataclass
class Option:
    """A config option."""

    value: Any = field(init=False)
    default: Any
    is_valid: Callable[[Any], bool]
    error_msg: str

    def __post_init__(self):
        self.value = self.default


def action_with_key(key: str, keyset: Dict[str, list]) -> Optional[str]:
    """Returns the *action* in *keyset* having key *key* or ``None`` if there's no
    such action.
    """
    for action, (k, *_) in keyset.items():
        if k == key:
            return action


def get_log_function(level) -> Callable[[str], None]:
    def log(msg: str) -> None:
        if logging.VERBOSE is None:  # logging not yet initialized
            notify.notify(msg, notify_level, "config", verbose=verbose)
        else:
            logging.log(msg, _logger, log_level, "config", verbose=verbose)

    notify_level = getattr(notify, level)
    log_level = getattr(_logging, level)
    verbose = level == "INFO"

    return log


def init_config() -> None:
    """Initializes user configuration."""
    for var, level in (("error", "ERROR"), ("info", "INFO"), ("warn", "WARNING")):
        globals()[var] = get_log_function(level)

    if user_config_file:
        load_config(user_config_file)
    elif xdg_config_file:
        load_xdg_config()

    for keyset in context_keys.values():
        for action in keyset.values():
            action[3:] = (True, True)  # "shown", "enabled"
    context_keys["global"]["Config"][3] = False  # Till the config menu is implemented
    expand_key[3] = False  # "Key bar" action should be hidden

    reconfigure_tui(_context_keys)


def load_config(config_file: str) -> None:
    """Loads a user config file."""

    def revert_context_update(
        keyset: Dict[str, list], prev_keyset: Dict[str, list]
    ) -> None:
        for properties, prev_properties in zip(keyset.values(), prev_keyset.values()):
            properties[:] = prev_properties

    try:
        with open(config_file) as f:
            config = json.load(f)
    except Exception as e:
        error(f"Failed to load {config_file!r} ({type(e).__name__}: {e}).")
        return

    keys = config.pop("keys", None)

    for name, value in config.items():
        try:
            option = config_options[name]
        except KeyError:
            warn(f"Unknown option {name!r} (in {config_file!r}).")
        else:
            if option.is_valid(value):
                option.value = value
            else:
                value_repr = "null" if value is None else repr(value)
                value_type_name = "null" if value is None else type(value).__name__
                error(
                    f"Invalid type/value for {name!r}; {option.error_msg} "
                    f"(got: {value_repr} of type {value_type_name!r})."
                )
                option_repr = "null" if option.value is None else repr(option.value)
                info(f"Using former value: {option_repr}.")
    if keys:
        prev_context_keys = deepcopy(context_keys)

        # Globals first...
        g_keyset = context_keys["global"]
        g_update = keys.pop("global", None)
        if g_update:
            prev_g_keyset = deepcopy(g_keyset)
            if not update_context("global", g_keyset, g_update, config_file):
                revert_context_update(g_keyset, prev_g_keyset)
                return

        # Then navigation...
        # Also going through if unupdated to detect conflicts between the unpdated keys
        # and global keys
        nav_update = keys.pop("navigation", {})
        prev_nav = deepcopy(nav)
        if not update_context("navigation", nav, nav_update, config_file):
            g_update and revert_context_update(g_keyset, prev_g_keyset)
            revert_context_update(nav, prev_nav)
            return

        # Done before updating other context keys so that actions having keys
        # conflicting with those for naviagation in the same context can be detected
        update_context_nav(context_keys, nav)
        navi.update((key, nav_action) for nav_action, (key, _) in nav.items())

        # Then other context actions.
        # Also going through unupdated contexts to detect conflicts between the
        # unpdated keys and navigation or global keys
        for context, keyset in context_keys.items():
            update = keys.pop(context, {})
            if not update_context(context, keyset, update, config_file):
                for keyset, prev_keyset in zip(
                    context_keys.values(), prev_context_keys.values()
                ):  # Includes the global context
                    revert_context_update(keyset, prev_keyset)
                revert_context_update(nav, prev_nav)
                navi.update((key, nav_action) for nav_action, (key, _) in nav.items())
                return
        for context in keys:
            warn(f"Unknown context {context!r} (in {config_file!r}).")


def load_xdg_config() -> None:
    """Loads user config files according to the XDG Base Directories spec."""
    for config_dir in reversed(os.environ.get("XDG_CONFIG_DIRS", "/etc").split(":")):
        config_file = path.join(config_dir, "term_image", "config.json")
        if (
            # The XDG Base Dirs spec states that relative paths should be ignored
            path.abspath(config_dir) == config_dir
            and path.isfile(config_file)
        ):
            load_config(config_file)

    if path.isfile(xdg_config_file):
        load_config(xdg_config_file)


def reconfigure_tui(
    old_context_keys: Optional[Dict[str, Dict[str, list]]] = None
) -> None:
    """Updates aspects of the TUI to use the current config option values and
    keybindings.
    """
    from . import logging
    from .tui.keys import change_key
    from .tui.widgets import expand, image_grid, notif_bar, pile

    command = urwid.command_map._command_defaults.copy()
    urwid.command_map._command = {
        nav[action][0]: command[key] for key, action in _navi.items()
    }

    if old_context_keys:
        for context, keyset in context_keys.items():
            old_keyset = old_context_keys[context]
            for action, (key, *_) in keyset.items():
                old_key = old_keyset[action][0]
                if old_key != key:
                    try:
                        change_key(context, old_key, key)
                    except KeyError:  # e.g navigation keys in "image-grid"
                        pass

    expand_or_collapse = expand.original_widget.text[0]
    expand.original_widget.set_text(f"{expand_or_collapse} [{expand_key[1]}]")

    if not logging.QUIET:
        if pile.contents[-1][0] is notif_bar:
            pile.contents.pop()
        if config_options.max_notifications:
            pile.contents.append(
                (notif_bar, ("given", config_options.max_notifications))
            )

    image_grid.cell_width = config_options.cell_width


def store_config(config_file: str) -> None:
    """Writes current config to a file."""
    config = {
        name: option.value
        for name, option in config_options.items()
        if option.value != option.default
    }

    modified_keys = {}
    modified_nav = {
        action: properties
        for _properties, (action, properties) in zip(_nav.values(), nav.items())
        if properties != _properties
    }
    if modified_nav:
        modified_keys["navigation"] = modified_nav
    for _keyset, (context, keyset) in zip(_context_keys.values(), context_keys.items()):
        context_nav = context_navs[context]
        keys = {}
        for _properties, (action, properties) in zip(_keyset.values(), keyset.items()):
            # Exclude context navigation actions and of course, unmodified actions
            if action not in context_nav and properties[:2] != _properties[:2]:
                keys[action] = properties[:2]  # Remove description and state
        if keys:  # Exclude contexts with only navigation actions
            modified_keys[context] = keys
    if modified_keys:
        config["keys"] = modified_keys

    err = None
    try:
        try:
            os.makedirs(path.dirname(config_file) or ".", exist_ok=True)
        except (FileExistsError, NotADirectoryError):
            err = "one of the parents is not a directory"
        else:
            with open(config_file, "w") as f:
                json.dump(config, f, indent=4)
    except Exception as e:
        err = f"{type(e).__name__}: {e}"
    if err:
        error(f"Failed to write user config to {config_file!r} ({err}).")


def update_context(
    context: str, keyset: Dict[str, list], update: Dict[str, list], config_file: str
) -> bool:
    """Updates the *keyset* of context *context* with *update*.

    Returns ``True`` if successful, otherwise ``False``.
    """

    def try_fallback(default: bool = False) -> bool:
        """Sets the fallback (old) or default key for the current action if not already
        assigned to another action.

        Returns ``True`` if successful, otherwise ``False``.
        """
        if default:
            default_keyset = _nav if context == "navigation" else _context_keys[context]
            _key = default_keyset[action][0]
            if _key == keyset[action][0]:
                info("... default key is the same as the current/former key.")
                error("... unable to find an unassigned fallback.")
                return False
        else:
            _key = keyset[action][0]
            if _key == key:
                info("... former key is the same as the new key.")
                return try_fallback(default=True)
        in_global = _key in global_
        in_assigned = _key in assigned
        fallback = "default" if default else "former"

        if in_global or in_assigned:
            if in_global:
                error(
                    f"... {fallback} key {_key!r} already assigned to global action "
                    f"{global_[_key]!r}."
                )
            elif in_assigned:
                _action = assigned[_key]
                error(
                    f"... {fallback} key {_key!r} already assigned to action "
                    f"{_action!r} "
                    + (
                        f"(derived from 'navigation::{context_nav[_action]}') "
                        if _action in context_nav
                        else ""
                    )
                    + "in the same context."
                )
            if default:
                error("... unable to find an unassigned fallback.")
                return False
            return try_fallback(default=True)

        assigned[_key] = action
        if default:
            keyset[action][:2] = default_keyset[action][:2]
        info(f"... using {fallback} key {_key!r}.")

        return True

    global_ = (
        set()
        if context == "global"
        else {key: action for action, (key, *_) in context_keys["global"].items()}
    )
    context_nav = context_navs.get(context, {})
    assigned = {keyset[action][0]: action for action in keyset.keys() - update.keys()}
    # Must include all context nav actions and they should override normal actions
    # using the same key since they have been updated earlier.
    assigned.update({keyset[action][0]: action for action in context_nav.keys()})

    failed = True
    for action, (key, *_) in keyset.items():
        context_action = f"'{context}::{action}'"
        try:
            properties = update.pop(action)
        except KeyError:
            conflict_msg = f"Key conclict with {context_action}..."
            if key in global_:
                error(conflict_msg)
                error(
                    f"... current key {key!r} already assigned to global action "
                    f"{global_[key]!r}."
                )
                if not try_fallback(default=True):
                    break
            elif assigned[key] != action:  # Has been assigned to a nav action earlier
                _action = assigned[key]
                error(conflict_msg)
                error(
                    f"... current key {key!r} already assigned to action {_action!r} "
                    f"(derived from 'navigation::{context_nav[_action]}') "
                    "in the same context."
                )
                if not try_fallback(default=True):
                    break
            continue

        if action in context_nav:
            warn(
                f"{context_action} should be updated via 'navigation::"
                f"{context_nav[action]}' (in {config_file!r})."
            )
            continue

        if not (
            isinstance(properties, list)
            and len(properties) == 2
            and all(isinstance(x, str) for x in properties)
        ):
            error(
                f"The properties ({properties!r}) of {context_action} are not in the "
                f"correct format (in {config_file!r})..."
            )
            if not try_fallback():
                break
            continue

        key, symbol = properties
        conflict_msg = (
            f"Key conclict with {context_action} (updated in {config_file!r})..."
        )

        if key not in _valid_keys:
            error(f"Invalid key {key!r} for {context_action} (in {config_file!r})...")
            if not try_fallback():
                break
        elif key in global_:
            error(conflict_msg)
            error(
                f"... new key {key!r} already assigned to global action "
                f"{global_[key]!r}."
            )
            if not try_fallback():
                break
        elif key in assigned:
            _action = assigned[key]
            error(conflict_msg)
            error(
                f"... new Key {key!r} already assigned to action {_action!r} "
                + (
                    f"(derived from 'navigation::{context_nav[_action]}') "
                    if _action in context_nav
                    else ""
                )
                + "in the same context."
            )
            if not try_fallback():
                break
        else:
            assigned[key] = action
            keyset[action][:2] = (key, symbol)
    else:
        failed = False
        for action in update:
            warn(
                f"Unknown action {action!r} in context {context!r} "
                f"(in {config_file!r})."
            )
    if failed:
        error(
            f"Unable to update with keybindings from {config_file!r} due to the "
            "previous error... Changes reverted."
        )

    return not failed


def update_context_nav(
    context_keys: Dict[str, Dict[str, list]],
    nav_update: Dict[str, list],
) -> None:
    """Updates keys and symbols of navigation actions in all contexts
    in *context_keys*.
    """
    for context, keyset in context_keys.items():
        context_nav = context_navs[context]
        for action, properties in keyset.items():
            if action in context_nav:
                properties[:2] = nav_update[context_nav[action]]


_logger = _logging.getLogger(__name__)
error: Callable[[str], None] = None
info: Callable[[str], None] = None
warn: Callable[[str], None] = None

user_config_file = None
xdg_config_file = path.join(
    os.environ.get("XDG_CONFIG_HOME", path.join(path.expanduser("~"), ".config")),
    "term_image",
    "config.json",
)

_valid_keys = {*bytes(range(32, 127)).decode(), *urwid.escape._keyconv.values(), "esc"}
_valid_keys.update(
    {
        f"{modifier} {key}"
        for key in (
            *(f"f{n}" for n in range(1, 13)),
            "delete",
            "end",
            "home",
            "up",
            "down",
            "left",
            "right",
        )
        for modifier in ("ctrl", "shift", "shift ctrl")
    }
)
_valid_keys.update(
    {
        f"ctrl {key}"
        for key in (
            *map(chr, range(ord("a"), ord("z") + 1)),
            "page up",
            "page down",
        )
    }
)
_valid_keys.difference_update({None, "ctrl c", "ctrl z"})

# For users and documentation
valid_keys = sorted(
    _valid_keys,
    key=lambda s: (
        chr(127 + len(s.rsplit(" ", 1)[-1]))  # group by main key
        + s.rsplit(" ", 1)[-1].lower()  # group both cases of alphabetical keys together
        + s.rsplit(" ", 1)[-1]  # sort alphabetical keys by case
        + chr(127 + len(s))  # sort by length within a group of the same main key
    ),
)
for key in ("page up", "page down"):
    valid_keys.remove(key)
    valid_keys.remove("ctrl " + key)
valid_keys.extend(("page up", "ctrl page up", "page down", "ctrl page down"))

config_options = {
    "anim cache": Option(
        100,
        lambda x: isinstance(x, int) and x > 0,
        "must be an integer greater than zero",
    ),
    "cell ratio": Option(
        None,
        lambda x: x is None or isinstance(x, float) and x > 0.0,
        "must be `null` or a float greater than zero",
    ),
    "cell width": Option(
        30,
        lambda x: isinstance(x, int) and 30 <= x <= 50 and not x % 2,
        "must be an even integer between 30 and 50 (both inclusive)",
    ),
    "checkers": Option(
        None,
        lambda x: x is None or isinstance(x, int) and x >= 0,
        "must be `null` or a non-negative integer",
    ),
    "getters": Option(
        4,
        lambda x: isinstance(x, int) and x > 0,
        "must be an integer greater than zero",
    ),
    "grid renderers": Option(
        1,
        lambda x: isinstance(x, int) and x >= 0,
        "must be a non-negative integer",
    ),
    "log file": Option(
        path.join("~", ".term_image", "term_image.log"),
        lambda x: isinstance(x, str) and is_writable(x),
        "must be a string containing a writable/creatable file path",
    ),
    "max notifications": Option(
        2,
        lambda x: isinstance(x, int) and 0 <= x <= 5,
        "must be an integer between 0 and 5 (both inclusive)",
    ),
    "max pixels": Option(
        2**22,  # 2048x2048
        lambda x: isinstance(x, int) and x > 0,
        "must be an integer greater than zero",
    ),
    "multi": Option(
        True,
        lambda x: isinstance(x, bool),
        "must be a boolean",
    ),
    "query timeout": Option(
        QUERY_TIMEOUT,
        lambda x: isinstance(x, float) and x > 0.0,
        "must be a float greater than zero",
    ),
    "style": Option(
        "auto",
        lambda x: x in {"auto", "block", "iterm2", "kitty"},
        "must be one of 'auto', 'block', 'iterm2', 'kitty'",
    ),
    "swap win size": Option(
        False,
        lambda x: isinstance(x, bool),
        "must be a boolean",
    ),
}
config_options = ConfigOptions(config_options)

_nav = {
    "Left": ["left", "\u25c0"],
    "Up": ["up", "\u25b2"],
    "Right": ["right", "\u25b6"],
    "Down": ["down", "\u25bc"],
    "Page Up": ["page up", "PgUp"],
    "Page Down": ["page down", "PgDn"],
    "Home": ["home", "Home"],
    "End": ["end", "End"],
}
_navi = {key: nav_action for nav_action, (key, _) in _nav.items()}

# {<context>: {<action>: [<key>, <symbol>, <desc>, <visibility>, <state>], ...}, ...}
# <visibility> and <state> are added later in `init_config()`.
_context_keys = {
    "global": {
        "Config": ["C", "\u21e7C", "Open configuration menu"],
        "Help": ["f1", "F1", "Show this help menu"],
        "Quit": ["q", "q", "Exit term-image"],
        "Key Bar": [".", ".", "Expand/Collapse key bar"],
    },
    "menu": {
        "Open": ["enter", "\u23ce", "Open the selected item"],
        "Prev": ["up", "", "Select the next item on the list"],
        "Next": ["down", "", "Select the previous item on the list"],
        "Back": ["backspace", "\u27f5 ", "Return to the previous directory"],
        "Delete": ["d", "d", "Delete selected image"],
        "Switch Pane": ["tab", "\u21b9", "Switch to image pane"],
        "Page Up": ["page up", "", "Jump up one page"],
        "Page Down": ["page down", "", "Jump down one page"],
        "Top": ["home", "", "Jump to the top of the list"],
        "Bottom": ["end", "", "Jump to the bottom of the list"],
    },
    "image": {
        "Prev": ["left", "", "Move to the previous image"],
        "Next": ["right", "", "Move to the next image"],
        "Force Render": [
            "F",
            "\u21e7F",
            "Force an image, with more pixels than the set maximum, to be displayed",
        ],
        "Maximize": ["f", "f", "Maximize the current image"],
        "Delete": ["d", "d", "Delete current image"],
        "Switch Pane": ["tab", "\u21b9", "Switch to list pane"],
    },
    "image-grid": {
        "Open": ["enter", "\u23ce", "Maximize the selected image"],
        "Up": ["up", "", "Move cursor up"],
        "Down": ["down", "", "Move cursor down"],
        "Left": ["left", "", "Move cursor left"],
        "Right": ["right", "", "Move cursor right"],
        "Switch Pane": ["tab", "\u21b9", "Switch to list pane"],
        "Size-": ["-", "-", "Decrease grid cell size"],
        "Size+": ["+", "+", "Increase grid cell size"],
        "Page Up": ["page up", "", "Jump up one page"],
        "Page Down": ["page down", "", "Jump down one page"],
        "Top": ["home", "", "Jump to the top of the grid"],
        "Bottom": ["end", "", "Jump to the bottom of the grid"],
    },
    "full-image": {
        "Restore": ["esc", "\u238b", "Exit maximized view"],
        "Prev": ["left", "", "Move to the previous image"],
        "Next": ["right", "", "Move to the next image"],
        "Force Render": [
            "F",
            "\u21e7F",
            "Force an image, with more pixels than the set maximum, to be displayed",
        ],
        "Delete": ["d", "d", "Delete current image"],
    },
    "full-grid-image": {
        "Back": ["esc", "\u238b", "Back to grid view"],
        "Force Render": [
            "F",
            "\u21e7F",
            "Force an image, with more pixels than the set maximum, to be displayed",
        ],
    },
    "confirmation": {
        "Confirm": ["enter", "\u23ce", ""],
        "Cancel": ["esc", "\u238b", ""],
    },
    "overlay": {
        "Close": ["esc", "\u238b", ""],
        "Up": ["up", "", "Scroll up"],
        "Down": ["down", "", "Scroll down"],
        "Page Up": ["page up", "", "Scroll up one page"],
        "Page Down": ["page down", "", "Scroll down one page"],
        "Top": ["home", "", "Jump to the top"],
        "Bottom": ["end", "", "Jump to the bottom"],
    },
}

nav = deepcopy(_nav)
navi = deepcopy(_navi)
context_navs = {
    context: {action: navi[key] for action, (key, *_) in keyset.items() if key in navi}
    for context, keyset in _context_keys.items()
}
update_context_nav(_context_keys, _nav)  # Update symbols
context_keys = deepcopy(_context_keys)
expand_key = context_keys["global"]["Key Bar"]
